/*
 *  arch/ppc/mm/init.c
 *
 *  Teleworld Customer Device version, loosly derived from the
 *  code indicated below.
 *    Copyright (C) 1998 Teleworld, Inc. All Rights Reserved.
 *
 *----------
 *
 *  PowerPC version 
 *    Copyright (C) 1995-1996 Gary Thomas (gdt@linuxppc.org)
 *
 *  Adapted for the Power Macintosh by Paul Mackerras.
 *    Copyright (C) 1996 Paul Mackerras (paulus@cs.anu.edu.au)
 *
 *  Derived from "arch/i386/mm/init.c"
 *    Copyright (C) 1991, 1992, 1993, 1994  Linus Torvalds
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 *
 */

#include <linux/config.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/head.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/ptrace.h>
#include <linux/mman.h>
#include <linux/mm.h>
#include <linux/swap.h>
#include <linux/stddef.h>
#include <linux/init.h>
#include <asm/prom.h>
#include <asm/io.h>
#include <asm/mmu_context.h>
#include <asm/pgtable.h>
#include <asm/mmu.h>
#include <asm/ppcops.h>
#include <asm/px_hw.h>


#define MMU_PAGE_SIZE	4096

int next_mmu_context = 1;
int NTLB_WIRED = 0;
extern pgd_t swapper_pg_dir[];

extern void die_if_kernel(char *,struct pt_regs *,long);
extern void show_net_buffers(void);
unsigned long dynomem;
int mem_init_done;
void squoosh_data_cache(void);

/*
 * BAD_PAGE is the page that is used for page faults when linux
 * is out-of-memory. Older versions of linux just did a
 * do_exit(), but using this instead means there is less risk
 * for a process dying in kernel mode, possibly leaving a inode
 * unused etc..
 *
 * BAD_PAGETABLE is the accompanying page-table: it is initialized
 * to point to BAD_PAGE entries.
 *
 * ZERO_PAGE is a special page that is used for zero-initialized
 * data and COW.
 */
unsigned long empty_bad_page_table;

extern int get_memconfig(void);
int PromMemConfig;
int PromVideoConfig;

pte_t * __bad_pagetable(void)
{
	memset((void *)empty_bad_page_table, 0, PAGE_SIZE);
	return (pte_t *) empty_bad_page_table;
}

unsigned long empty_bad_page;

pte_t __bad_page(void)
{
	memset((void *)empty_bad_page, 0, PAGE_SIZE);
	return pte_mkdirty(mk_pte(empty_bad_page, PAGE_SHARED));
}

void show_mem(void)
{
	int i,free = 0,total = 0,reserved = 0;
	int shared = 0;

	printk("Mem-info:\n");
	show_free_areas();
	printk("Free swap:       %6dkB\n",nr_swap_pages<<(PAGE_SHIFT-10));
	i = max_mapnr;
	while (i-- > 0) {
		total++;
		if (PageReserved(mem_map+i))
			reserved++;
		else if (!mem_map[i].count)
			free++;
		else
			shared += mem_map[i].count-1;
	}
	printk("%d pages of RAM\n",total);
	printk("%d free pages\n",free);
	printk("%d reserved pages\n",reserved);
	printk("%d pages shared\n",shared);
	show_buffers();
#ifdef CONFIG_NET
	show_net_buffers();
#endif
}

extern unsigned long free_area_init(unsigned long, unsigned long);

/*
 * paging_init() sets up the page tables - in fact we've already done this.
 */
unsigned long __init paging_init(unsigned long start_mem, unsigned long end_mem)
{
	/*
	 * Grab some memory for bad_page and bad_pagetable to use.
	 */
	empty_bad_page = start_mem;
	empty_bad_page_table = start_mem + PAGE_SIZE;
	start_mem += 2 * PAGE_SIZE;

	/* note: free_area_init uses its second argument
	   to size the mem_map array. */
	start_mem = free_area_init(start_mem, end_mem);
	return start_mem;
}

int calc_vmalloc_start(void)
{

	if (PromMemConfig > 16)
		return(0x80000000 + (32*1024*1024));
	else 
		return(0x80000000 + (16*1024*1024));
#if 0
	switch (PromMemConfig) {
	case 16:
		return(0x80000000 + (16*1024*1024));
	case 32:
		return(0x80000000 + (32*1024*1024));
	default:
	case 8:
		return(0x80000000 + (16*1024*1024));
	}
#endif
}

void __init mem_init(unsigned long start_mem, unsigned long end_mem)
{
	unsigned long a;
	unsigned long codesize, total_mem;
	int np;
	extern char etext[], _stext[];
	extern struct mm_struct init_mm;

	end_mem &= PAGE_MASK;
	high_memory = (void *) end_mem;
	max_mapnr = MAP_NR(high_memory);
	num_physpages = max_mapnr;	/* RAM is assumed contiguous */

	/* mark usable pages in the mem_map[] */
	start_mem = PAGE_ALIGN(start_mem);
	dynomem = start_mem;
	np = 0;
	printk("mem_init: start_mem = 0x%lX, end_mem = 0x%lX\n",
	       start_mem, end_mem);
	for (a = start_mem; a < end_mem ; a += PAGE_SIZE) {
		clear_bit(PG_reserved, &mem_map[MAP_NR(a)].flags);
		mem_map[MAP_NR(a)].count = 1;
		free_page(a);
		np++;
	}
	/* initialize region for kernel virtual allocation */
	init_mm.mmap->vm_start = VMALLOC_START;
	init_mm.mmap->vm_end = VMALLOC_START + (np * PAGE_SIZE);
	total_mem = end_mem - PAGE_OFFSET;
	codesize = PAGE_ALIGN((unsigned)etext) - (unsigned)_stext;
	a = nr_free_pages << PAGE_SHIFT;
	printk("Memory: %luk/%luk available (%luk kernel code, %luk data)\n",
	       a >> 10, total_mem >> 10,
	       codesize >> 10, (total_mem - a - codesize) >> 10);
	printk("Kernel virtual start=0x%lx, end=0x%lx\n", init_mm.mmap->vm_start,
		init_mm.mmap->vm_end);

	mem_init_done = 1;
}

void free_initmem(void)
{
	unsigned long a;
	struct page *p;
	int num_freed = 0;
	extern char __init_begin, __init_end;

	/*
	 * We can free up the PROM memory area as well.
	 */
        a = (unsigned long) PAGE_OFFSET;
	for (p = &mem_map[0]; p < &mem_map[MAP_NR(KERNELBASE)]; p++) {
		clear_bit(PG_reserved, &p->flags);
		p->count = 1;
		free_page(a);
                a += PAGE_SIZE;
		num_freed++;
	}
	printk("%d pages of PROM memory freed\n", num_freed);
	num_freed = 0;
	a = (unsigned long) &__init_begin;
	for (; a < (unsigned long)&__init_end; a+= PAGE_SIZE) {
		clear_bit(PG_reserved, &mem_map[MAP_NR(a)].flags);
		(mem_map+MAP_NR(a))->count = 1;
		free_page(a);
		num_freed++;
	}
	printk("%d pages of unused kernel memory freed\n",
	       num_freed);
		
}

void si_meminfo(struct sysinfo *val)
{
	int i;

	i = max_mapnr;
	val->totalram = 0;
	val->sharedram = 0;
	val->freeram = nr_free_pages << PAGE_SHIFT;
	val->bufferram = buffermem;
	while (i-- > 0)  {
		if (PageReserved(mem_map+i))
			continue;
		val->totalram++;
		if (!mem_map[i].count)
			continue;
		val->sharedram += mem_map[i].count-1;
	}
	val->totalram <<= PAGE_SHIFT;
	val->sharedram <<= PAGE_SHIFT;
	return;
}

/* Kernel MMU setup & lowest level hardware support */

/*
 * We "wire" a TLB entry permanently for the kernel.
 * XXX If bigger than 16M physical, need to do multiple TLB entries
 *
 * The memory management structure of the 403 series is a bit
 * complicated, since there are three different translation
 * mechanisms that all interact, and have different cache policies:
 * TLB mapping (virtual mode), zones (virtual mode) and boundary
 * registers (real mode). Additionally, the Bus Interface Unit (BIU)
 * strips the top bit of all real addresses it gets, meaning that
 * there are two addresses for any particular real location. This
 * bit, however, *is* carried throughout all other calculations
 * (real page numbers in the TLB, for instance). This means that 
 * it's pretty easy to get cache synonyms (multiple copies of the
 * same line at different virtual addresses in the cache) if things
 * aren't managed carefully.
 *
 * Therefore, I've adopted the following set of policies:
 *
 * 1. The kernel is always mapped at 0x8000000 on up, both in
 *    virtual and real mode, so there is never any translation
 *    to worry about. User code occupies the lower 2GB starting
 *    at virtual 0.
 * 2. Kernel pages are mapped with PID 0 (meaning the pages are
 *    shared among all processes); each unique process has a unique
 *    PID. User-level access to kernel pages is restricted using
 *    the zone register. Kernel pages are loaded into zone 0, 
 *    which is set for no user access. User pages are always loaded
 *    into zone 1, which is set for both user and kernel access.
 *    Other zones are currently not used.
 * 3. Using the cacheability control registers (ICCR and DCCR),
 *    the kernel address space is set for caching, but no other
 *    addresses are. This means that in real mode, access to
 *    the lower 2GB of address space is non-cached. Kernel DRAM
 *    pages are cached in virtual mode; this setup means that whether
 *    the kernel is running in virtual or real mode, it uses identical
 *    addresses for DRAM, and these access are always cached. This is
 *    important because the exception handlers run in real mode,
 *    and get executed alot, so we want their accesses cached.
 * 4. Non-cached access to kernel DRAM (buffers, for instance), must
 *    occur through a separate TLB mapping that does non-cached access.
 *    By convention, this area starts at 0xc0000000, continuing for the
 *    size of DRAM. IO devices are mapped using the TLB non-cached,
 *    at arbitraily chosen addresses above 0x80000000.
 * 5. Various page sizes are used in the TLB for efficiency for
 *    kernel DRAM and IO pages; user pages are always 4K in size,
 *    unless the memory is special use (e.g., memory pools). The first
 *    NTLB_WIRED entries are reserved for permanent TLB entries (e.g.,
 *    kernel DRAM); otherwise, a random replacement strategy is used
 *    based on the CPU timebase.
 * 6. The kernel spends as little time in real mode as possible. The
 *    general philosophy of exception handlers is 1) save state, 
 *    2) go virtual (into C). This may need to be revisited for certain
 *    handlers (TLB miss, for example).
 *
 * -- jmb
 */

#define DRAM1	(1 * 1024 * 1024)
#define DRAM2	(2 * 1024 * 1024)
#define DRAM4	(4 * 1024 * 1024)
#define DRAM8	(8 * 1024 * 1024)
#define DRAM12	(12 * 1024 * 1024)
#define DRAM16	(16 * 1024 * 1024)


/* Signature should be 'TiVo'.  If so, params points to boot params */
void
MMU_init(int signature, char *params)
{
#define CONFIG_TCD_MEMLOG 1
#if CONFIG_TCD_MEMLOG
    	int initmemlog = 0;
#endif

	void tlbWirekernel(unsigned long va, unsigned long pa, int, int, int);
	void flush_instruction_cache(void);
	int i;
	unsigned long kaddr = (KERNELBASE & ~0xffffff);
	extern void s403_init(void);
	extern void save_boot_params(char *params);

	asm("tlbia");
	asm("sync;isync");

	if (signature == 'TiVo')
		save_boot_params(params);
        PromMemConfig = get_memconfig();
        PromVideoConfig = get_videoconfig();

#if CONFIG_TCD_MEMLOG
	if( PromMemConfig == 32 )
	{
	    int len = strlen( params ) - 5;

	    for( i = 0; i < len; ++i )
	    {
		if( params[ i ] == 'm' )
		{
		    if( strncmp( &params[i], "memlog", 6 ) == 0 )
		    {
			initmemlog = 1;			
	    
			PromMemConfig = 16;	
		    }		    
		}		
	    }
	}	
#endif
	if (PromMemConfig > 16)
		i = 32;
	else 
		i = 16;

	switch (i) {
	case 16:
	default:
		tlbWirekernel(kaddr, kaddr, TLBHI_PGSZ_16M, NTLB_WIRED++, 1);
		break;
	case 32:
		tlbWirekernel(kaddr, kaddr, TLBHI_PGSZ_16M, NTLB_WIRED++, 1);
		tlbWirekernel(kaddr+DRAM16, kaddr+DRAM16, TLBHI_PGSZ_16M, NTLB_WIRED++, 1);
		break;
	}
#if 0
/* ??? Why is this here ??? scs */
        tlbWirekernel(0x90000000, kaddr, TLBHI_PGSZ_16M, NTLB_WIRED++, 0);
#endif

	/* place for kernel page tables */
	init_task.tss.pg_tables = (unsigned long *)swapper_pg_dir;

	/* Zone 0 - kernel (above 0x80000000), zone 1 - user */
	mtspr(SPRN_ZPR, 0x2aaaaaaa);
	mtspr(SPRN_DCWR, 0);		/* all caching is write-back */

	/* flush all data cache sets */
	for (i = 0; i < 256; i++) {
		asm("dcbf 0,%0" : : "r" (i << 4));
		asm("dccci 0,%0" : : "r" (i << 4));
	}
	mtspr(SPRN_DCCR, 0x00008000);

	flush_instruction_cache();
	mtspr(SPRN_ICCR, 0x00008000);

	/* on-board serial port */
	tlbWirekernel(IOBASE_SERIAL, TCD_IO_BASE_SERIAL, TLBHI_PGSZ_4K, NTLB_WIRED++, 0);

	/* Ethernet chip */
	tlbWirekernel(IOBASE_8390, TCD_IO_BASE_8390, TLBHI_PGSZ_4K, NTLB_WIRED++, 0);

#ifndef CONFIG_TCD_EVAL
	/* Configure TL750 */
	if (PromVideoConfig == 6) {

#define CACHE750

#ifdef CACHE750
		/* NOTE: This is hardwired to cached 750 DRAM */
		tlbWirekernel(IOBASE_INPUT, IOBASE_INPUT,
			TLBHI_PGSZ_4M, NTLB_WIRED++, 1);
		/* NOTE: If you cache the 750 DRAM, set MS control */
#define TL750_BURST_ENABLE (1<<19)
		MsWrite(MS_CTL_REG, MsRead(MS_CTL_REG) | TL750_BURST_ENABLE);
#else
		tlbWirekernel(IOBASE_INPUT, IOBASE_INPUT,
			TLBHI_PGSZ_4M, NTLB_WIRED++, 0);
#endif
	} else {
		/* input section is 32MB*/
		tlbWirekernel(IOBASE_INPUT, TCD_IO_BASE_INPUT,
			TLBHI_PGSZ_16M, NTLB_WIRED++, 0);
		tlbWirekernel(IOBASE_INPUT+DRAM16, TCD_IO_BASE_INPUT+DRAM16,
			TLBHI_PGSZ_16M, NTLB_WIRED++, 0);
	}
#if 1
#define XXX 0xf9000000
	tlbWirekernel(XXX, XXX, TLBHI_PGSZ_16M, NTLB_WIRED++, 0);
#endif

	/* other section is 8MB */
	tlbWirekernel(_IOBASE_OTHER_, _IOBASE_OTHER_, TLBHI_PGSZ_1M, NTLB_WIRED++, 0);
	tlbWirekernel(_IOBASE_OTHER_+DRAM1, _IOBASE_OTHER_+DRAM1, TLBHI_PGSZ_4M, NTLB_WIRED++, 0);
#else
	/* UART */
	tlbWirekernel(IOBASE_UART, TCD_IO_BASE_UART, TLBHI_PGSZ_4K, NTLB_WIRED++, 0);
#endif
#ifdef CONFIG_TCD_MEMLOG
	/* Map the upper 16MB to 0xA0000000 accessable to all processes */
	if( initmemlog )
	{	    
		unsigned long tlbhi = 0xA0000000 | TLBHI_PGSZ_16M | TLBHI_VALID;
		unsigned long tlblo = 0x01000000 | TLBLO_Z_USER | TLBLO_WR;
		unsigned long opid = mfspr(SPRN_PID);
		mtspr(SPRN_PID, 0);
		asm("tlbwe %0,%1,1" : : "r" (tlblo), "r" (NTLB_WIRED));
		asm("tlbwe %0,%1,0" : : "r" (tlbhi), "r" (NTLB_WIRED));
		asm("isync");
		mtspr(SPRN_PID, opid);
		++NTLB_WIRED;

		*(long*)0x01000000 = 0x0;
	}	
#endif
	asm("sync;isync");
}

void *
ioremap(unsigned long addr, unsigned long size)
{
	panic("there is no need to call ioremap!");
}

/*
 * TLB flushing:
 *
 *  - flush_tlb_all() flushes all processes TLBs
 *  - flush_tlb_mm(mm) flushes the specified mm context TLB's
 *  - flush_tlb_page(vma, vmaddr) flushes one page
 *  - flush_tlb_range(mm, start, end) flushes a range of pages
 *
 */

void tlbFlush(void);
void tlbFlush1(unsigned long va, int pid);

/*
 * The context counter has overflowed.
 * We set mm->context to NO_CONTEXT for all mm's in the system.
 * We assume we can get to all mm's by looking as tsk->mm for
 * all tasks in the system.
 */
void
mmu_context_overflow(void)
{
	extern int _tlbpidwrap;
	struct task_struct *tsk;
	int nr;

	printk(KERN_DEBUG "mmu_context_overflow\n");
	for (nr = 0; nr < NR_TASKS; ++nr) {
		tsk = task[nr];
		if (tsk && tsk->mm && (tsk->mm->context != 0))
			tsk->mm->context = NO_CONTEXT;
	}
	next_mmu_context = 1;
	tlbFlush();
	_tlbpidwrap++;

	/* make sure current always has a context */
	current->mm->context = next_mmu_context++;
	set_context(current->mm->context);
}

/*
 * Flush all tlb/hash table entries (except perhaps for those
 * mapping RAM starting at PAGE_OFFSET, since they never change).
 */
void
flush_tlb_all(void)
{
	tlbFlush();
}

/*
 * Flush all the (user) entries for the address space described
 * by mm.  We can't rely on mm->mmap describing all the entries
 * that might be in the hash table.
 */
void
flush_tlb_mm(struct mm_struct *mm)
{
	mm->context = NO_CONTEXT;
	if (mm == current->mm) {
		get_mmu_context(current);
		set_context(current->mm->context);
	}
}

/*
 * for each page addr in the range, call MMU_invalidate_page().
 * For a large range, it's quicker just to give the mm a new context.
 */
void
flush_tlb_range(struct mm_struct *mm, unsigned long start, unsigned long end)
{
	flush_tlb_mm(mm);
}

/*
 * Cache flushing operations
 */

void flush_instruction_cache(void) {
	unsigned long p = KERNELBASE & PAGE_MASK;
	unsigned long stop = p + (16 * 1024);
	for (; p <= stop; p += 16)
		asm("iccci 0,%0" : : "r" (p) : "r0");
	asm("sync;isync");
}

void flush_data_cache(void) {
	unsigned long p = KERNELBASE & PAGE_MASK;
	unsigned long stop = p + (8 * 1024);
	for (; p <= stop; p += 16)
		asm("dccci 0,%0" : : "r" (p) : "r0");
	asm("sync;isync");
}

void store_cache_range(unsigned long p, unsigned long stop) {
        p =       p & 0xFFFFFFF0;
        stop = stop | 0x0000000F;
	for (; p <= stop; p += 16) {
		asm("dcbf 0,%0" : : "r" (p));
		asm("icbi  0,%0" : : "r" (p));
	}
}

void flush_cache_all(void) { }

void flush_cache_mm(struct mm_struct *mm) { }

void flush_cache_range(struct mm_struct *mm, unsigned long start, unsigned long stop) { }

void flush_cache_page(struct vm_area_struct *vma, unsigned long page) { }

/*
 * Flush kernel physical addresses. Doesn't need to do anything special
 * because virtual and real are the same for physical pages.
 */
void flush_page_to_ram(unsigned long page) {
	unsigned long p = page & PAGE_MASK;
	unsigned long stop = p + PAGE_SIZE;
	for (; p < stop; p += 16) {
		asm("dcbf 0,%0" : : "r" (p));
		asm("icbi 0,%0" : : "r" (p));
	}
	asm("sync;isync");
}

/*
 * Invalidate any data in the caches for a dma operation.
 */
void dma_invalidate_cache_range(unsigned long p, unsigned long stop) {
        p =       p & 0xFFFFFFF0;
        stop = stop | 0x0000000F;
	for (; p <= stop; p += 16) {
		asm("dcbi 0,%0" : : "r" (p));
	}
        asm("sync");

}

void dma_flush_cache_range(unsigned long p, unsigned long stop) {
        p =       p & 0xFFFFFFF0;
	for (; p < stop; p += 16) {
		asm("dcbf 0,%0" : : "r" (p));
	}
	asm("sync");
}

/*
 * This is a gross misnaming.  The flush_data_cache() routine does
 * not flush the cache lines - it forcibly invalidates the lines.
 * It's supposed to be used only during system startup - if used
 * at other times it'll probably cause data loss.
 */

void dma_flush_cache_all() {
	flush_instruction_cache();
	flush_data_cache();
}

/*
 * Force out the current contents of the data cache, by loading
 * data from the kernel code area.  Specific to PowerPC 403, may
 * not work correctly on other models.  dplatt 9/6/2000
 */

void squoosh_data_cache() {
	unsigned long p = KERNELBASE & PAGE_MASK;
	long entries = (8 * 1024 / 16);
	unsigned long flags;
        unsigned long dummy = 0;
	save_flags(flags);
	cli();
	do {
	   asm("lwz %0,0(%1)" : : "r" (dummy), "r" (p) );
	   p += 16;
	} while (entries-- > 0);
	asm("sync");
        restore_flags(flags);
}

/*
 * Scan the dcache, and force out each tagged address.  This is
 * designed for the PowerPC 403GCX and may not work on other CPUs.
 */

void untag_data_cache() {
	unsigned long p = KERNELBASE & PAGE_MASK;
        unsigned long pIndex = p & 0x00000FFF;
        int lines = 256;
        unsigned long tagMask = 0xFFFFF000;
        unsigned long cdbcrA = 0x10;
        unsigned long cdbcrB = 0x11;
        unsigned long tagA = 0, tagB = 0;
	unsigned long flags;
        if (pIndex != 0) {
	   panic("PAGE_MASK isn't pageful!");
	}
	save_flags(flags);
	cli();
        do {
	   asm ("mtspr 0x3D7,%0" : : "r" (cdbcrA));
	   asm ("dcread %0,0,%1" : : "r" (tagA), "r" (p));
	   tagA = (tagA & tagMask) | pIndex;
	   asm ("mtspr 0x3D7,%0" : : "r" (cdbcrB));
	   asm ("dcread %0,0,%1" : : "r" (tagB), "r" (p));
	   tagB = (tagB & tagMask) | pIndex;
	   asm("dcbf 0,%0" : : "r" (tagA));
	   asm("dcbf 0,%0" : : "r" (tagB));
	   p += 16;
	   pIndex += 16;
	} while (lines-- > 0);
        asm("sync");
        restore_flags(flags);
}


/*
 * Dangling FPU stuff.
 */
void giveup_fpu() {}



