/*
 * OS debug functions.
 *
 * Copyright (C) 1998 Teleworld, Inc.
 * All Rights Reserved.
 */

#include <linux/config.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/head.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/ptrace.h>
#include <linux/mman.h>
#include <linux/mm.h>
#include <linux/swap.h>
#include <linux/stddef.h>
#include <linux/vmalloc.h>
#include <linux/fs.h>
#include <linux/pagemap.h>
#include <asm/page.h>
#include <asm/prom.h>
#include <asm/io.h>
#include <asm/mmu_context.h>
#include <asm/pgtable.h>
#include <asm/mmu.h>
#include <asm/ppcops.h>
#include <asm/delay.h>
#include "nonstdio.h"
#include <linux/tivodefs.h>
#include <asm/px_hw.h>

#ifdef CONFIG_XMON

extern void printf(const char *fmt, ...);
unsigned long strtoul(char *, char **, int);
void backtrace(unsigned int);
extern void (*xmon_fault_handler)(struct pt_regs *);
extern void handle_fault(struct pt_regs *);
extern u_int bus_error_jmp[100];
extern unsigned long dynomem;
extern void *high_memory;
extern struct inode *first_inode;
extern int nr_inodes;
extern struct inode *get_inode_hashlist(struct inode *);
extern void excprint(struct pt_regs *);
extern struct pt_regs _xdebug_faultregs;
extern void prregs(struct pt_regs *);
int xmon_setjmp(u_int *buf);
#define setjmp xmon_setjmp
#define longjmp xmon_longjmp

static void
_xos_mem_map(int pageno) {
	unsigned int mask;
	int i;
	struct page *p = mem_map + pageno;
	printf("struct page: %p (%d)\n", p, pageno);
	printf(" next=%p, prev=%p, inode=%p, offset=%d, next_hash=%p\n",
		p->next, p->prev, p->inode, p->offset, p->next_hash);
	printf(" count=%d, dirty=%d, age=%d, wait=%p, prev_hash=%p\n",
		p->count, p->dirty, p->age, p->wait, p->prev_hash);
	printf(" buffers=%p, swap_unlock_entry=%d, map_nr=%d\n",
		p->buffers, p->swap_unlock_entry, p->map_nr);
	printf(" flags (");
	for (i = 0, mask = 1; i < 32; i++, mask <<= 1) {
            if (p->flags & mask) switch (i) {
		case PG_locked: printf("locked "); break;
		case PG_error: printf("error "); break;
		case PG_referenced: printf("referenced "); break;
		case PG_uptodate: printf("uptodate "); break;
		case PG_free_after: printf("free_after "); break;
		case PG_decr_after: printf("decr_after "); break;
		case PG_swap_unlock_after: printf("swap_unlock_after "); break;
		case PG_DMA: printf("DMA "); break;
		case PG_reserved: printf("reserved "); break;
		}
	}
	printf(")\n");
}

static int
xos_mem_map(int ac, char **av, struct pt_regs *regs) {
	static int pageno = 0;
	static int epage = 0;
	int i;
	if (pageno == 0) {
		pageno = MAP_NR(dynomem);
		epage = MAP_NR(high_memory);
		printf("there are %d pages of dynamic memory available\n",
			epage - pageno);
	}
	if (ac == 1) {
		_xos_mem_map(++pageno);
		return 0;
	}
	pageno = (int)strtoul(av[1], 0, 0);
	if (pageno < 0 || pageno >= MAP_NR(high_memory)) {
	    	printf("invalid starting page number\n");
		return 0;
	}
	if (ac == 2) {
		_xos_mem_map(pageno);
                return 0;
        }
	epage = (int)strtoul(av[2], 0, 0);
	if (epage < 0 || epage >= MAP_NR(high_memory)) {
	    	printf("invalid ending page number\n");
		return 0;
	}
	for (i = pageno; i < epage; i++)
	    	_xos_mem_map(i);
	return 0;
}

static int
xos_mem(int ac, char **av, struct pt_regs *regs) {
    	extern void show_mem(void);
	show_mem();
	return 0;
}

#if 0
static void
_xos_vm_struct(struct vm_struct *vm) {
	printf("struct vm_struct: %p\n", vm);
	printf("\taddr=%p, size=%d, next=%p flags=%x\n", vm->addr,
		vm->size, vm->next, vm->flags);
}

static int
xos_vm_struct(int ac, char **av, struct pt_regs *regs) {
	static struct vm_struct *vm = 0;
	if (ac == 1) {
		if (vm == 0) {
			printf("no vm_struct entry yet set for showing\n");
			return 0;
		}
		_xos_vm_struct(vm);
		return 0;
	}
	if (ac >= 2) {
		vm = (struct vm_struct *) strtoul(av[1], 0, 0);
		_xos_vm_struct(vm);
		return 0;
	}
	return 0;
}
#endif

static void
_xos_vm_area(struct vm_area_struct *vma, int pages) {
	int mask = 1;
        int nopagecount = 0;
	printf("struct vm_area_struct: %p, %d total virtual pages\n", vma,
		(vma->vm_end - vma->vm_start) / PAGE_SIZE);
	printf("  vm_mm=%p, vm_start=%p, vm_end=%p, vm_offset=%d, vm_pte=%p\n",
		vma->vm_mm, vma->vm_start, vma->vm_end, vma->vm_offset, vma->vm_pte);
	printf("  vm_page_prot=%x, vm_next=%p, avl_left=%p, avl_right=%p\n",
		vma->vm_page_prot, vma->vm_next, vma->vm_avl_left, vma->vm_avl_right);
	printf("  flags=(");
	for (; mask; mask <<= 1) switch(vma->vm_flags & mask) {
		case VM_READ: printf("r "); break;
		case VM_WRITE: printf("w "); break;
		case VM_EXEC: printf("x "); break;
		case VM_SHARED: printf("s "); break;
		case VM_MAYREAD: printf("mr "); break;
		case VM_MAYWRITE: printf("mw "); break;
		case VM_MAYEXEC: printf("mx "); break;
		case VM_MAYSHARE: printf("ms "); break;
		case VM_GROWSDOWN: printf("down "); break;
		case VM_GROWSUP: printf("up "); break;
		case VM_SHM: printf("shm "); break;
		case VM_DENYWRITE: printf("~w "); break;
		case VM_EXECUTABLE: printf("exc "); break;
		case VM_LOCKED: printf("locked "); break;
		case VM_IO: printf("io "); break;
	}
	printf(")\n");
	{
		pte_t *findPTE(struct mm_struct *, unsigned long);
	    	unsigned long addr = vma->vm_start;
		int nvalid = 0;
		int nwrite = 0;
		int ntotal = 0;
		while (addr < vma->vm_end) {
		    	pte_t *pte = findPTE(vma->vm_mm, addr);
		    	ntotal++;
                        if (pte) {
                            if (nopagecount > 2) {
                                printf("... (times %d)\n",
                                        nopagecount - 2);
                            }
                            nopagecount = 0;
                            if (pages) {
                                _xos_mem_map(MAP_NR(pte_page(*pte)));
                            }
                            nvalid++;
                            if (pte_write(*pte))
                                    nwrite++;
                        }
                        else if (pages) {
                            nopagecount++;
                            if (nopagecount < 3) {
                                printf("no page present for 0x%8lx\n", addr);
                            } 
                        }
			addr += PAGE_SIZE;
		}
                if (nopagecount > 2) {
                    printf("... (times %d)\n",
                            nopagecount - 2);
                }
		printf("  %d total pages, %d valid pages, %d writable pages\n",
			ntotal, nvalid, nwrite);
	}
}

static int
xos_vm_area(int ac, char **av, struct pt_regs *regs) {
	static struct vm_area_struct *vm = 0;
	if (ac == 1) {
		if (vm == 0) {
			printf("no vm_area_struct entry yet set for showing\n");
			return 0;
		}
		_xos_vm_area(vm, 0);
		return 0;
	}
	if (ac >= 2) {
		vm = (struct vm_area_struct *) strtoul(av[1], 0, 0);
                if (ac >= 3)
                    _xos_vm_area(vm, 1);
                else
                    _xos_vm_area(vm, 0);
                        return 0;
	}
	return 0;
}

static void
_xos_mm(struct mm_struct *mm) {
	printf("struct mm_struct: %p\n", mm);
	printf("  count=%d, pgd=%p, context=%d, mmap=%p, mmap_avl=%p\n",
		mm->count, mm->pgd, mm->context, mm->mmap, mm->mmap_avl);
	printf("  start_code=%p, end_code=%p, start_data=%p, end_data=%p\n",
		mm->start_code, mm->end_code, mm->start_data, mm->end_data);
    printf("  arg_start=%p, arg_end=%p, env_start=%p, env_end=%p\n",
        mm->arg_start, mm->arg_end, mm->env_start, mm->env_end);
}

static void
_xos_walk_vm(struct vm_area_struct *vma) {
	_xos_vm_area(vma, 0);
	if (vma->vm_avl_left)
		_xos_walk_vm(vma->vm_avl_left);
	if (vma->vm_avl_right)
		_xos_walk_vm(vma->vm_avl_right);
}

static int
xos_mm(int ac, char **av, struct pt_regs *regs) {
	static struct mm_struct *mm = 0;
	int idx;
	if (ac == 1) {
		if (!mm) {
			if (current == 0) {
				printf("no current task\n");
				return 0;
		    	}
			else
			    	mm = current->mm;
		}
		return 0;
	}
	idx = 1;
	if (av[idx][0] >= '0' && av[idx][0] <= '9') {
		mm = (struct mm_struct *) strtoul(av[idx], 0, 0);
		idx++;
	}
	if (ac > idx) {
	    	if (!strcmp(av[idx], "all")) {
			_xos_mm(mm);
			_xos_walk_vm(mm->mmap_avl);
		}
		else
		    	printf("invalid dump option\n");
	}
	else
		_xos_mm(mm);
	return 0;
}

static void
_xos_buffer(struct buffer_head *buf) {
	int i;
	static char *buf_types[NR_LIST] = {"CLEAN","LOCKED","LOCKED1","DIRTY"};
	printf("struct buffer_head %p: dev %d, rdev %d, block %d, sector %d\n",
		buf, buf->b_dev, buf->b_rdev, buf->b_blocknr, buf->b_rsector);
	printf("  state=(");
	for (i = 0; i < 32; i++) {
		if (buf->b_state & (1 << i))
			switch (i) {
			case BH_Uptodate: printf("uptodate "); break;
			case BH_Dirty: printf("dirty "); break;
			case BH_Lock: printf("lock "); break;
			case BH_Req: printf("req "); break;
			case BH_Touched: printf("touched "); break;
			case BH_Has_aged: printf("aged "); break;
			case BH_Protected: printf("protected "); break;
			case BH_FreeOnIO: printf("freeOnIO "); break;
			}
	}
	printf("), count %d, size %d, page %d\n", buf->b_count, buf->b_size,
			MAP_NR(buf->b_data));
	printf("  reg %p, next %p, prev %p, list %s, wait %p\n",
		buf->b_reqnext, buf->b_next, buf->b_prev,
		buf_types[buf->b_list],
		buf->b_wait);
}

int
xos_buffers(int ac, char **av, struct pt_regs *regs) {
	int lidx = 1;
	int idx = 1;
	struct buffer_head *bh;
	extern struct buffer_head *lru_list[];

	while (idx < ac && !(av[idx][0] >= '0' && av[idx][0] <= '9')) {
		if (!strcmp(av[idx], "clean"))
			lidx = 0;
		else if (!strcmp(av[idx], "locked"))
			lidx = 1;
		else if (!strcmp(av[idx], "locked1"))
			lidx = 2;
		else if (!strcmp(av[idx], "dirty"))
			lidx = 3;
		else {
			printf("invalid buffer list type\n");
			return 0;
		}
		idx++;
	}
	if (idx < ac && (av[idx][0] >= '0' && av[idx][0] <= '9')) {
		struct buffer_head *eb = (struct buffer_head *) strtoul(av[idx], 0, 0);
		_xos_buffer(eb);
		return 0;
	}
	bh = lru_list[lidx];
	if (!bh) {
		printf("list is empty\n");
		return 0;
	}
	do {
		_xos_buffer(bh);
		bh = bh->b_next_free;
	} while (bh != lru_list[lidx]);
	return 0;
}

struct free_area_struct {
	struct page *next;
	struct page *prev;
	unsigned int * map;
};

static void
_xos_free_pages(void) {
	int i;
	printf("free pages: %d free, %d swap\n", nr_free_pages, nr_swap_pages);
	for (i = 0; i < 6; i++) {
		struct page *p;
		extern struct free_area_struct free_area[];
		struct free_area_struct *fsp = &free_area[i];
		printf("  freelist %d (%d page(s))\n", i, i+1);
		for (p = fsp->next; p != (struct page *)fsp; p = p->next) {
			printf("    page %d (%p), count %d\n", p->map_nr, p,
				p->count);
		}
	}
}

static int
xos_free_pages(int ac, char **av, struct pt_regs *regs) {
	extern void show_free_areas(void);
	if (ac < 2) {
		show_free_areas();
		return 0;
	}
	if (!strcmp(av[1], "details")) {
	    _xos_free_pages();
	    return 0;
	}
	else {
		printf("unrecognized argument\n");
		return 0;
	}
	return 0;
}

static void
_xos_current(struct task_struct *task) {
	int vector = 0;
        int mask;
	if (task->tss.regs) {
		vector = task->tss.regs->trap;
	}
        else
                vector = 0;
	printf("task %p (pid %d, \"%s\") vector %p rtpri %d counter %d ",
		task, task->pid, task->comm, vector, task->rt_priority, task->counter);
        switch (task->state) {
            case TASK_RUNNING: printf("R"); break;
            case TASK_INTERRUPTIBLE: printf("I"); break;
            case TASK_UNINTERRUPTIBLE: printf("U"); break;
            case TASK_ZOMBIE: printf("Z"); break;
            case TASK_STOPPED: printf("T"); break;
            case TASK_SWAPPING: printf("W"); break;
        }
        printf("\n");
	printf("    tss=%p, mm=%p, spid=%d, kregs=%p, regs=%p\n", &task->tss, task->mm,
		task->mm->context, task->tss.ksp+16, task->tss.regs);
        printf("    signal %x, blocked %x, flags (", task->signal, task->blocked);
        for ( mask = 1; mask < PF_ONSIGSTK; mask <<= 1 ) {
                if ( ! ( mask & task->flags ) )
                    continue;
                switch ( mask ) {
                    case PF_ALIGNWARN: printf("ALIGN "); break;
                    case PF_PTRACED: printf("TRACE "); break;
                    case PF_TRACESYS: printf("TRACESYS "); break;
                    case PF_FORKNOEXEC: printf("FKNOEX "); break;
                    case PF_SUPERPRIV: printf("SUPRIV "); break;
                    case PF_DUMPCORE: printf("CORE "); break;
                    case PF_SIGNALED: printf("SIG "); break;
                    case PF_STARTING: printf("START "); break;
                    case PF_EXITING: printf("EXIT "); break;
                }
        }
        printf(")\n");
}

static int
xos_current(int ac, char **av, struct pt_regs *regs) 
{
    int idx;
    struct task_struct *stask = 0;
    int nr;
    if (ac > 1) {
        unsigned long taskNum = strtoul(av[1], 0, 0);
        if (taskNum >= 0x80000000) {
            stask = (struct task_struct *) strtoul(av[1], 0, 0);
        } else {
            for (nr = 0; nr < NR_TASKS; nr++) {
                stask = task[nr];
                if (stask != NULL && stask->pid == taskNum) break;
            }
            if (nr == NR_TASKS) {
                printf("Task number %d unknown\n", taskNum);
                return 0;
            }
        }
    }
    else {
        stask = current;
    }
    _xos_current(stask);
    idx = 2;
    while (idx < ac) {
	struct mm_struct *saved = current->mm;
	pgd_t *spgdir = 0;
	if (current != stask) {
		// Reset the mm to something useful for backtracing.
		current->mm = stask->mm;
		mtspr(SPRN_PID, current->mm->context);
		spgdir = (pgd_t *) current->tss.pg_tables;
		current->tss.pg_tables = (unsigned long *) stask->mm->pgd;
	}

	if (!strcmp(av[idx], "kbt") ||
	    !strcmp(av[idx], "ebt") ||
	    !strcmp(av[idx], "bt")) {
		if (current != stask)
			backtrace((unsigned)((struct pt_regs *)stask->tss.ksp));
		else
			backtrace(0);
	} else if (!strcmp(av[idx], "mm")) {
	    _xos_mm(stask->mm);
	    _xos_walk_vm(stask->mm->mmap_avl);
	} else {
	    printf("unknown keyword \"%s\"\n", av[idx]);
	}
	idx++;

	if (current != stask) {
		// Restore current->mm
		current->mm = saved;
		mtspr(SPRN_PID, current->mm->context);
		current->tss.pg_tables = (unsigned long *)spgdir;
	}
    }
    return 0;
}

static int
xos_tasks(int ac, char **av, struct pt_regs *regs) 
{
    int dokbt = 0;
    int nr;
    if (ac > 1) {
        if (strcmp(av[1], "kbt") != 0) {
            printf("invalid option\n");
            return 0;
        }
        dokbt = 1;
    }
    for (nr = 0; nr < NR_TASKS; nr++) {
        struct task_struct *tsk = task[nr];
        if (tsk) {
            _xos_current(tsk);
            if (dokbt) {
		if (current != tsk) {
			// Reset the mm to something useful for backtracing.
			pgd_t *spgdir;
			struct mm_struct *saved = current->mm;
			current->mm = tsk->mm;
			mtspr(SPRN_PID, current->mm->context);
			spgdir = (pgd_t *) current->tss.pg_tables;
			current->tss.pg_tables = (unsigned long *) tsk->mm->pgd;

			backtrace((unsigned)(struct pt_regs *)tsk->tss.ksp);

			// Restore current->mm
			current->mm = saved;
			mtspr(SPRN_PID, current->mm->context);
			current->tss.pg_tables = (unsigned long *)spgdir;
		}
		else
			backtrace(0);
            }
        }
    }
    return 0;
}

static void
_xos_dumppte(pte_t *pte) {
	printf("%p  ", pte);
	if (!pte) {
		printf("invalid entry\n");
		return;
	}
	if (pte_none(*pte)) {
		printf("no entry\n");
		return;
	}
	if (!pte_present(*pte)) {
		printf("no page present\n");
		return;
	}
	printf("phys %p NR %d (", pte_page(*pte), MAP_NR(pte_page(*pte)));
	{
		unsigned int mask = 0x1;
		int i;
		for (i = 0; i < 12; i++, mask <<= 1) {
			if (mask & *((unsigned int *)pte)) switch (mask) {
				case _PAGE_PRESENT: printf("present "); break;
				case _PAGE_USER: printf("user "); break;
				case _PAGE_RW: printf("rw "); break;
				case _PAGE_GUARDED: printf("guard "); break;
				case _PAGE_COHERENT: printf("coherent "); break;
				case _PAGE_NO_CACHE: printf("nocache "); break;
				case _PAGE_WRITETHRU: printf("writethrough "); break;
				case _PAGE_DIRTY: printf("dirty "); break;
				case _PAGE_ACCESSED: printf("ref "); break;
				case _PAGE_HWWRITE: printf("hwwrite "); break;
			}
		}
	}
	printf(")\n");
	return;
}

static int
xos_pte(int ac, char **av, struct pt_regs *regs) {
	static int npte = 10;
	static unsigned long addr = 0;
	struct mm_struct *mm;
	int i;

	if (ac != 4) {
	    	printf("usage: pte <MMSTRUCT> <VADDR> <NPAGE>\n");
		return 0;
	}
	mm = (struct mm_struct *) strtoul(av[1], 0, 0);
	addr = strtoul(av[2], 0, 0);
	npte = strtoul(av[3], 0, 0);
	for (i = 0; i < npte; i++) {
		pte_t *findPTE(struct mm_struct *, unsigned long);
		pte_t *tpte = findPTE(mm, addr);
		_xos_dumppte(tpte);
		addr += PAGE_SIZE;
	}
	return 0;
}

static int
xos_xlate(int ac, char **av, struct pt_regs *regs) {
	struct vm_area_struct *vma;
	struct mm_struct *mm;
	pgd_t *pgd;
	pmd_t *pmd;
	pte_t *pte;
	unsigned long badaddr;

	if (ac <= 1) {
		printf("insufficient arguments\n");
		return 0;
	}
	badaddr = strtoul(av[1], 0, 0);
	if (ac >= 3)
		mm = (struct mm_struct *)strtoul(av[2], 0, 0);
	else if (badaddr >= VMALLOC_START)
                mm = task[0]->mm;
        else
		mm = current->mm;
	if (!mm) {
		printf("no current mm struct\n");
		return 0;
	}
	vma = find_vma(current->mm, badaddr);
	if (!vma) {
		printf("no matching vm area found (0x%lx)\n", badaddr);
		return 0;
	}
	pgd = pgd_offset(current->mm, badaddr);
	if (pgd_none(*pgd)) {
		printf("no page directory entry found (0x%lx)\n", badaddr);
		return 0;
	}
	pmd = pmd_offset(pgd, badaddr);
	if (pmd_none(*pmd)) {
		printf("no page middle entry found (0x%lx)\n", badaddr);
		return 0;
	}
	pte = pte_offset(pmd, badaddr);
	if (pte_none(*pte)) {
		printf("no page table entry found (0x%lx)\n", badaddr);
		return 0;
	}
	if (!pte_present(*pte)) {
		printf("page not present , PTE = %p\n", *pte);
		return 0;
	}
	printf("PGD: %p PMD %p PTE: %p, *pte = %p\n", pgd, pmd, pte, *pte);
	_xos_dumppte(pte);
	return 0;
}

static int
xos_irq(int ac, char **av, struct pt_regs *regs) {
    extern void dump_cbuf(void);
    dump_cbuf();
    return 0;
}

#if 0
static int
_xos_check_file(struct file *f) {
    int e = 0;
    /*
    if (f->f_count) {
        if (!f->f_inode && f->f_op) {
            printf("warning: file 0x%lx has no inode\n", f);
            e++;
        }
    }
    */
    return e;
}

static int
_xos_check_files(void) {
    u_int sbus[100];
    struct file *f;
    struct file *follow = 0;
    int i;
    int e = 0;
    void (*sfault)(struct pt_regs *);

    memcpy(sbus, bus_error_jmp, sizeof(u_int) * 100);
    sfault = xmon_fault_handler;
    if (setjmp(bus_error_jmp)) {
        memcpy(bus_error_jmp, sbus, sizeof(u_int) * 100);
        xmon_fault_handler = sfault;
        printf("File check fault, file 0x%lx\n", follow);
        return 1;
    }
    xmon_fault_handler = handle_fault;

    for (f = first_file, i = 0; i < nr_files; f = f->f_next, i++) {
        e += _xos_check_file(f);
        follow = f;
    }

    memcpy(bus_error_jmp, sbus, sizeof(u_int) * 100);
    xmon_fault_handler = sfault;
    return e;
}
#endif

static void
_xos_file(struct file *f) {
    printf("%p: pos %10lu flags 0x%04x count %2u inode 0x%8lx next 0x%lx\n",
            f,
            (unsigned int)(f->f_pos & 0xffffffff),
            f->f_flags, f->f_count, f->f_inode, f->f_next);
}

static int
xos_file(int ac, char **av, struct pt_regs *regs) {
    struct file *f;
    if (ac == 1) {
        int i;
        printf("File table dump, %d file structures present\n", nr_files);
        for (f = first_file, i = 0; i < nr_files; f = f->f_next, i++) {
            _xos_file(f);
        }
        return 0;
    }
    f = (struct file *) strtoul(av[1], 0, 0);
    _xos_file(f);
    return 0;
}

static int
_xos_scanmap(struct inode *ip) {
    int i = 0;
    struct page *page = mem_map + MAP_NR(dynomem);
    struct page *end = mem_map + MAP_NR(high_memory);
    for (; page < end; page++)
        if (page->inode == ip)
            i++;
    return i;
}

static int
_xos_check_inode(struct inode *ip, int verbose) {
    int j;
    int e = 0;
    if (ip->i_pages) {
        struct page *page;
        int i = ip->i_nrpages;
        for (page = ip->i_pages; page != 0; page = page->next) {
            printk("page list %p\r", page);
            if (page->inode != ip) {
                printf("warning: inode 0x%lx and page %d don't point at each other\n",
                        ip, MAP_NR(page));
                return e;
            }
            else if (page->next_hash != 0) {
                struct page *hash = *page_hash(page->inode, page->offset);
                struct page *follow = 0;
                while (hash != 0) {
                    printk("page hash %p -> %p\r", follow, hash);
                    follow = hash;
                    if (hash == page)
                        break;
                    hash = hash->next_hash;
                }
                if (!hash) {
                    printf("warning: inode 0x%lx, page %d not hashed\n",
                            ip, MAP_NR(page));
                    e++;
                }
            }
            i--;
        }
        if (i != 0) {
            printf("warning: inode 0x%lx page count, list size differ <%d -> %d>\n",
                    ip, ip->i_nrpages, ip->i_nrpages - i);
            e++;
        }
        if ((j = _xos_scanmap(ip)) != ip->i_nrpages) {
            printf("warning: inode 0x%lx page references (%d) != inode count (%d)\n",
                    ip, j, ip->i_nrpages);
            e++;
        }
    } else {
        struct page *page = mem_map + MAP_NR(dynomem);
        struct page *end = mem_map + MAP_NR(high_memory);
        for (; page < end; page++) {
            if (page->inode == ip) {
                struct page *pl;
                for (pl = ip->i_pages; pl != 0; pl = page->next) {
                    if (pl == page)
                        break;
                }
                if (pl == 0) {
                    printf("warning: inode 0x%lx page 0x%lx don't match\n",
                            ip, page);
                    e++;
                }
            }
        }
    }
    if (ip->i_count && ip->i_dev != 0) {
        struct inode *hip = get_inode_hashlist(ip);
        if (!hip) {
            printf("warning: inode 0x%lx hash list is empty!\n", ip);
            e++;
        }
        else {
            struct inode *follow = 0;
            for (; hip != 0; hip = hip->i_hash_next) {
                printk("inode hash %p -> %p\r", follow, hip);
                follow = hip;
                if (hip == ip)
                    break;
            }
            if (!hip) {
                printf("warning: inode 0x%lx not on proper hash list\n", ip);
                e++;
            }
        }
    }
    return e;
}

#if 0
static int _xos_check_inodes(void) {
    struct inode *ip;
    int i;
    int e = 0;
    for (ip = first_inode, i = 0; i < nr_inodes; ip = ip->i_next, i++) {
        if (ip->i_count)
            e += _xos_check_inode(ip, 0);
    }
    return e;
}

static int
xos_alloc(int ac, char **av, struct pt_regs *regs) {
    extern void _alloc_dump(unsigned long, int);
    int max = 0x7fffffff;
    if (ac == 1)
        _alloc_dump(0, max);
    else if (ac > 2)
        max = strtoul(av[2], 0, 0);
    _alloc_dump(strtoul(av[1], 0, 0), max);
    return 0;
}

#endif

static void
_xos_inode(struct inode *ip) {
    printf("0x%lx: dev 0x%04x ino %6ld count %3d flags 0x%04x hnext 0x%8lx (page 0x%lx)\n",
            ip, ip->i_dev, ip->i_ino, ip->i_count, ip->i_flags,
            ip->i_hash_next, ip->i_pages);
}

static int
xos_inode(int ac, char **av, struct pt_regs *regs) {
    struct inode *ip;
    if (ac == 1) {
        int i;
        printf("Inode table dump, %d inode structures present\n", nr_inodes);
        for (ip = first_inode, i = 0; i < nr_inodes; ip = ip->i_next, i++) {
            if (ip->i_count)
                _xos_inode(ip);
            else if (ip->i_dirt || ip->i_lock)
                printf("0x%lx: inode is dirty or locked, but not in use\n");
        }
        return 0;
    }
    ip = (struct inode *) strtoul(av[1], 0, 0);
    _xos_inode(ip);
    if (ac > 2)
        _xos_check_inode(ip, 1);
    return 0;
}

static void
_xos_checkvma(struct task_struct *task, struct vm_area_struct *vma, unsigned long pageno) {
        pte_t *findPTE(struct mm_struct *, unsigned long);
        unsigned long addr = vma->vm_start;
        while (addr < vma->vm_end) {
                pte_t *pte = findPTE(vma->vm_mm, addr);
                if (pte && pte_present(*pte)) {
                    if (MAP_NR(pte_page(*pte)) == pageno)
                        printf("task 0x%lx: vmaddr 0x%lx points at %d\n", task,
                                addr, pageno);
                }
                addr += PAGE_SIZE;
        }
}

static void
_xos_walkvma(struct task_struct *task, struct vm_area_struct *vma, unsigned long pageno) {
	_xos_checkvma(task, vma, pageno);
	if (vma->vm_avl_left)
            _xos_walkvma(task, vma->vm_avl_left, pageno);
	if (vma->vm_avl_right)
            _xos_walkvma(task, vma->vm_avl_right, pageno);
}

static int
xos_fpage(int ac, char **av, struct pt_regs *regs) {
    int pageno;
    int nr;
    struct buffer_head *bh;
    extern struct buffer_head *lru_list[];
    extern int _alloc_checkfree(unsigned long);
    if (ac < 2) {
        printf("usage: os fpage <PAGENO>\n");
        return 0;
    }
    pageno = strtoul(av[1], 0, 0);
    for (nr = 0; nr < NR_TASKS; nr++) {
            struct task_struct *tsk = task[nr];
            if (tsk && tsk->mm->mmap_avl)
                    _xos_walkvma(tsk, tsk->mm->mmap_avl, pageno);
    }
#if 0
    if (_alloc_checkfree(pageno))
            printf("page %d is on the free list\n", pageno);
#endif
    for (nr = 0; nr < NR_LIST; nr++) {
        bh = lru_list[nr];
        if (!bh)
            continue;
        do {
            if (MAP_NR(bh) == pageno)
                printf("page %d contains buffer headers\n", pageno);
            if (MAP_NR(bh->b_data) == pageno)
                printf("page %d is pointed to by buffer header %p\n",
                        pageno, bh);
            bh = bh->b_next_free;
        } while (bh != lru_list[nr]);
    }
    return 0;
}

static void
_xos_wait(struct wait_queue *q) {
    struct wait_queue *p = q;
    printf("waiting tasks: ");
    while (p) {
        printf("%p ", p);
        p = p->next;
        if (p == q)
            break;
    }
    printf("\n");
}

static int
xos_wait(int ac, char **av, struct pt_regs *regs) {
    struct wait_queue *wait;
    if (ac < 2) {
        printf("usage: os wait <ADDR>\n");
        return 0;
    }
    wait = (struct wait_queue *) strtoul(av[1], 0, 0);
    _xos_wait(wait);
    return 0;
}

static int
xos_mod(int ac, char **av, struct pt_regs *regs) {
    extern void _module_dump(void);
    _module_dump();
    return 0;
}

static int
xos_guard(int ac, char **av, struct pt_regs *regs) {
    extern void dumpguards(void);
    dumpguards();
    return 0;
}

static int
xos_sched(int ac, char **av, struct pt_regs *regs) {
	extern struct task_struct init_task;
	struct task_struct *p;
	printf("Scheduler Run Queue:\n");
	for (p = init_task.next_run; p != &init_task; p = p->next_run) {
		printf("%p policy %s pri %3d rtpri %3d\n", p,
			(p->policy == SCHED_OTHER ? "other" :
			 (p->policy == SCHED_FIFO ? "fifo " :
			  (p->policy == SCHED_RR  ? "rr   " : "?????"))),
			p->priority, p->rt_priority);
	}
	return 0;
}

static int
xos_setmm(int ac, char **av, struct pt_regs *regs) {
    struct mm_struct *mm;
    struct mm_struct *mm2;
    extern struct mm_struct *_xdebug_restore_mm;
    if (ac != 2) {
        printf("usage: os setmm <MMADDR>\n");
        return 0;
    }
    if ((mm = (struct mm_struct *) strtoul(av[1], 0, 0)) == 0) {
        printf("bad mm address: %s\n", av[1]);
        return 0;
    }
    if (!_xdebug_restore_mm)
        _xdebug_restore_mm = current->mm;
    mm2 = current->mm;
    current->mm = mm;
    mtspr(SPRN_PID, current->mm->context);
    current->tss.pg_tables = (unsigned long *)mm->pgd;
    printf("Virtual space changed from %p to %p\n", mm2, mm);
    printf("The old virtual space will be restored on exit from the debugger\n");
    return 0;
}

#if redundant
static void
_xos_findva(struct task_struct *tsk, unsigned long addr) {
    struct vm_area_struct *vma = task->mm->mmap;
    addr &= PAGE_MASK;
    while (vma) {
        unsigned long va = vma->vm_start;
        unsigned long endva = vma->vm_end;
        while (va < endva) {
            pgd_t *pgd;
            pmd_t *pmd;
            pte_t *pte;
            pgd = pgd_offset(tsk->mm, va);
            if (pgd_none(*pgd))
                    goto NextIteration;
            pmd = pmd_offset(pgd, va);
            if (pmd_none(*pmd))
                    goto NextIteration;
            pte = pte_offset(pmd, va);
            if (pte_none(*pte))
                    goto NextIteration;
            if (!pte_present(*pte))
                    goto NextIteration;
            if (pte_page(*pte) == addr)
                printf("In task 0x%08lx, page 0x%08lx is VA 0x%08lx\n",
                    tsk, addr, va);
        NextIteration:
            va += PAGE_SIZE;
        }
        vma = vma->vm_next;
    }
}

static int
xos_findva(int ac, char **av, struct pt_regs *regs) {
        int nr;
        unsigned long addr;
	if (ac != 2) {
            printf("usage: findva <ADDR>\n");
            return 0;
        }
        addr = strtoul(av[1], 0, 0);
	for (nr = 0; nr < NR_TASKS; nr++) {
		struct task_struct *tsk = task[nr];
		if (tsk)
			_xos_findva(tsk, addr);
	}
	return 0;
}
#endif

#define SW_PR_ALL	0x1
#define SW_PR_INTR	0x2
#define SW_PR_BH	0x4
#define SW_PR_RQ	0x8
#define SW_PR_SW	0x10
#define SW_PR_SC	0x20
#define SW_PR_GEN	0x40
#define SW_PR_DRQ	0x80

static int
xos_sw(int ac, char **av, struct pt_regs *regs) {
	extern void _dump_swbuf(int, int);
	int flags = 0;
	int count = 0;
	int i;

        if (ac > 1 && !strcmp(av[1], "pid")) {
            if (ac != 3) {
                printf("pid option must be standalone\n");
                return 0;
            }
            _dump_swbuf(-strtoul(av[2], 0, 0), 0);
            return 0;
        }

	for (i = 1; i < ac; i++) {
	    if (!strcmp(av[i], "all"))
		flags |= SW_PR_ALL;
	    else if (!strcmp(av[i], "intr"))
		flags |= SW_PR_INTR;
	    else if (!strcmp(av[i], "bh"))
		flags |= SW_PR_BH;
	    else if (!strcmp(av[i], "rq"))
		flags |= SW_PR_RQ;
	    else if (!strcmp(av[i], "sw"))
		flags |= SW_PR_SW;
	    else if (!strcmp(av[i], "sc"))
		flags |= SW_PR_SC;
	    else if (!strcmp(av[i], "gen"))
		flags |= SW_PR_GEN;
	    else if (!strcmp(av[i], "drq"))
		flags |= SW_PR_DRQ;
	    else if (av[i][0] <= '9' && av[i][0] >= '0') {
		count = strtoul(av[i], 0, 0);
		break;
	    }
	    else {
		printf("usage: sw [[intr|bh|rq|drq|sw|sc|gen|all]|pid <PID>] [COUNT]\n");
		return 0;
	    }
	}
	if (flags == 0)
	    flags = -1 & ~0x1;
	else if (flags == SW_PR_ALL)
	    flags = -1;
	_dump_swbuf(count, flags);
	return 0;
}

static int
xos_swap(int ac, char **av, struct pt_regs *regs) {
	extern void dump_swaplist(void);
	dump_swaplist();
	return 0;
}


//////////////////////////////////////////////////////////////////////////////
//
// Routines for the "savecore" feature.  I tried to keep them as minimal as
// possible...
//
//////////////////////////////////////////////////////////////////////////////
#define MAC_PARTITION_MAGIC     0x504d

struct mac_partition {
    unsigned short  signature;      /* expected to be MAC_PARTITION_MAGIC */
    unsigned short  res1;
    unsigned long   map_count;      /* # blocks in partition map */
    unsigned long   start_block;    /* absolute starting block # of partition */
    unsigned long   block_count;    /* number of blocks in partition */
    char            name[32];       /* partition name */
    char            type[32];       /* string type description */
    unsigned long   data_start;     /* rel block # of first data block */
    unsigned long   data_count;     /* number of data blocks */
    unsigned long   status;         /* partition status bits */
    unsigned long   boot_start;
    unsigned long   boot_size;
    unsigned long   boot_load;
    unsigned long   boot_load2;
    unsigned long   boot_entry;
    unsigned long   boot_entry2;
    unsigned long   boot_cksum;
    char    processor[16];  /* identifies ISA of boot */
    /* there is more stuff after this that we don't need */
};

#define IDE_DEV_BSY         0x80
#define IDE_DEV_DRQ         0x08
#define IDE_DEV_ERROR       0x01

// Read a sector from disk 0.  Only for use by xos_savecore who resets the
// disk first.
static int
read_sector(unsigned long block, void *pDest) 
{
    unsigned long st;
    int i;
    unsigned short *pDestIter = (unsigned short *) pDest;

    // Initialize the 256 16bits words devivce information.
    MsWrite(MS_IDE_SECTOR_COUNT_REG, 0x1);
    MsWrite(MS_IDE_SECTOR_NUMBER_REG, block & 0xff);
    MsWrite(MS_IDE_CYLINDER_LOW_REG, (block >> 8) & 0xff);
    MsWrite(MS_IDE_CYLINDER_HIGH_REG, (block >> 16) & 0xff);
    MsWrite(MS_IDE_DEVICE_HEAD_REG, 0x40 | ((block >> 24) & 0xf));
    MsWrite(MS_IDE_STATUS_REG, 0x00000020);

    while ((st = MsRead(MS_IDE_ALT_STATUS_REG)) & IDE_DEV_BSY) ;

    if (st & IDE_DEV_ERROR) return 0;

    for (i = 0; i < 256; i++) {
        *pDestIter++ = MsRead(MS_IDE_DATA_REG) & 0xffff;
    }

    return 1;
}

// Write a sector to disk 0.  Only for use by xos_savecore who resets the disk
// first.
static int
write_sector(unsigned long block, void *pSource) 
{
    unsigned long st;
    unsigned int i;
    unsigned short *pSourceIter = (unsigned short *) pSource;

    MsWrite(MS_IDE_SECTOR_COUNT_REG, 0x1);
    MsWrite(MS_IDE_SECTOR_NUMBER_REG, block & 0xff);
    MsWrite(MS_IDE_CYLINDER_LOW_REG, (block >> 8) & 0xff);
    MsWrite(MS_IDE_CYLINDER_HIGH_REG, (block >> 16) & 0xff);
    MsWrite(MS_IDE_DEVICE_HEAD_REG, 0x40 | ((block >> 24) & 0xf));
    MsWrite(MS_IDE_STATUS_REG, 0x00000030);
    udelay(2000);

    while ((st = MsRead(MS_IDE_ALT_STATUS_REG)) & IDE_DEV_BSY) ;

    if (st & IDE_DEV_ERROR) return 0;

    while (!(st & IDE_DEV_DRQ)) {
        st = MsRead(MS_IDE_ALT_STATUS_REG);
    }

    for (i = 0; i < 256; i++) {
        while (MsRead(MS_IDE_ALT_STATUS_REG) & IDE_DEV_BSY) ;
        MsWrite(MS_IDE_DATA_REG, *pSourceIter);
        pSourceIter++;
    }
    while ((st = MsRead(MS_IDE_ALT_STATUS_REG)) & IDE_DEV_BSY) ;

    if (st & IDE_DEV_ERROR) return 0;

    return 1;
}


static int 
xos_savecore(int ac, char **av, struct pt_regs *regs) 
{
    unsigned char diskBlock[512];
    extern int PromMemConfig;
    struct tcdbootpage *pBootPage;
    int i;
    int memsize;
    int rootPartition, savePartition;
    int saveBlock;
    char *saveData;
    char *s;
    struct mac_partition *pPartition;

    printf("Resetting ide device -- continuing from debugger may not work\n");
    // Reset logic stolen from the prom.
    MsWrite(MS2_SYSTEM_REGISTER, MsRead(MS2_SYSTEM_REGISTER) | 0x8);
    udelay(10000);
    MsWrite(MS2_SYSTEM_REGISTER, MsRead(MS2_SYSTEM_REGISTER) & ~0x8);
    udelay(10000);

    // ide0 pio mode 4 clock 1 address setup, 2 active, 1 recovery
    // ide1 pio mode 0 clock 2 address setup, 8 active, 9 recovery
    // ide command mode 0 clock 2 address setup, 8 active, 9 recovery
    MsWrite(MS_IDE0_PIO_DATA_TIMING_REG,0xfffffd13);
    MsWrite(MS_IDE1_PIO_DATA_TIMING_REG,0xfffffd13);
    MsWrite(MS_IDE_PIO_COMMAND_TIMING_REG,0xfffffd13);

    // ide0 dma mode 2 clock 2 active, 2recovery
    // ide1 dma mode 0 clock 6 active, 6recovery
    MsWrite(MS_IDE0_MWDMA_TIMING_REG,0xfffffc12);
    MsWrite(MS_IDE1_MWDMA_TIMING_REG,0xfffffc12);

    // Reset device 0 (only drive we care about...)
    printf("Resetting drive 0...\n");
    MsWrite(MS_IDE_DEVICE_HEAD_REG, 0);
    udelay(20000);
    MsWrite(MS_IDE_STATUS_REG, 0x08);   // Reset
    MsWrite(MS_IDE_DEVICE_HEAD_REG, 0);
    udelay(10000);
    i = 0;
    while (MsRead(MS_IDE_STATUS_REG) & 0x80) {
        udelay(50000);
        if (i++ > 600) {     /* do it the old fashioned way */
            printf("Failed to reset ide interface\n");
            return 0;
        }
    }

    // Read boot sector.
    printf("Reading boot sector\n");
    if (!read_sector(0, diskBlock)) {
        printf("Failed to read boot sector\n");
        return 0;
    }

    pBootPage = (struct tcdbootpage *) diskBlock;
    if (pBootPage->signature != TIVO_BOOT_MAGIC) {
        printf("Invalid boot sector\n");
        return 0;
    }
    printf("disk bootparams = %s\n", pBootPage->bootParams);
    printf("boot partition = %d\n", pBootPage->primaryBootPartition);
    s = pBootPage->bootParams;
    rootPartition = 0;
    while (*s != 0) {
        if (strncmp(s, "root=/dev/hda", 13) == 0) {
            rootPartition = s[13] - '0';
            if (rootPartition != 4 && rootPartition != 7) {
                rootPartition = 0;
            }
            break;
        }
        s++;
    }
    if (rootPartition == 0) {
        printf("Failed to determine your root partition...  Aborting!\n");
        return 0;
    }
    if (rootPartition != pBootPage->primaryBootPartition+1) {
        printf("Your root partition isn't your boot partition + 1???\n");
        return 0;
    }
    savePartition = 11 - rootPartition;

    printf("Saving to /dev/hda%d\n", savePartition);
    // Read partition table.  Just for fun to make sure we know what we are
    // doing!

    if (!read_sector(savePartition, diskBlock)) {
        printf("Failed to read disk block %d\n", i);
        return 0;
    }
    pPartition = (struct mac_partition *) diskBlock;
    if (pPartition->signature != MAC_PARTITION_MAGIC) {
        printf("Cannot read partition %d header\n", savePartition);
    }
    printf("Partition %d:  Size %dM\n", savePartition,
            pPartition->block_count/2048);
    memsize = 2048*PromMemConfig;
    if (pPartition->block_count < memsize) {
        printf("Partition too small?!??\n");
        return 0;
    }
    saveBlock = pPartition->start_block;
    saveData = (char *) 0x80000000;
    printf("Writing %d 512 byte blocks starting at disk block %d\n", memsize,
            saveBlock);
    for (i = 0; i < memsize; i++) {
        write_sector(saveBlock, saveData);
        saveData += 512;
        saveBlock++;
        if (i % 100 == 0) {
            printf("%d\r", i);
        }
    }
    printf("\nDone\n");

    return 0;
}

static int xos_help(int, char **, struct pt_regs *);
#if 0
static int xos_check_opt(int, char **, struct pt_regs *);
#endif

static struct {
	char	*cmd;
	int	(*proc)(int ac, char **av, struct pt_regs *);
	char	*help;
} cmdtable[] = {
	{ "task",	xos_current,	"task [pid|ADDRESS]:<lasttask> [kbt ebt mm]" },
	{ "tasks",	xos_tasks,	"tasks [kbt]" },
	{ "page",	xos_mem_map,	"page [PAGENO]:<lastpageno> [EPAGE]:<lastepage>" },
	{ "mm",		xos_mm,		"mm [ADDRESS]:<lastaddr> [all]" },
	{ "mem",	xos_mem,	"mem" },
	{ "buf",	xos_buffers,	"buf ([clean|locked|dirty|locked1]|ADDRESS)" },
	{ "xlate",	xos_xlate,	"xlate ADDRESS" },
	{ "free_pages",	xos_free_pages,	"free_pages" },
	{ "pte",	xos_pte,	"pte <MM> <VADDR> <NPAGE>" },
        { "irq",        xos_irq,        "irq" },
        { "file",       xos_file,       "file [<ADDR>]" },
        { "inode",      xos_inode,      "inode [<ADDR>] [check]" },
        { "guard",      xos_guard,      "guard" },
        { "setmm",      xos_setmm,      "setmm <MMADDR>" },
        { "vma",        xos_vm_area,    "vma <ADDR> [pages]" },
        { "fpage",      xos_fpage,      "fpage <PAGENO>" },
        { "wait",       xos_wait,       "wait <ADDR>" },
        { "mod",        xos_mod,        "mod" },
	{ "runq",	xos_sched,	"runq" },
        { "help",       xos_help,       "help" },
	{ "sw",		xos_sw,		"sw [[intr|bh|rq|drq|sw|sc|gen|all] [COUNT]|pid <PID>]" },
	{ "swap",	xos_swap,	"swap" },
        { "savecore",   xos_savecore,   "savecore" },
	{ 0 },
};

static int
xos_help(int ac, char **av, struct pt_regs *regs) {
    int i;
	for (i = 0; cmdtable[i].cmd != 0; i++) {
            printf("os %s\n", cmdtable[i].help);
	}
        return 0;
}

int
xos(int ac, char **av, struct pt_regs *regs) {
	int i;

        // os with no args shouldn't crash...
        if (ac == 1) {
            xos_help(ac, av, regs);
            return 0;
        }

	if (setjmp(bus_error_jmp)) {
		printf("\ndebugger fault!\n");
                excprint(&_xdebug_faultregs);
                prregs(&_xdebug_faultregs);
		return 0;
	}
	xmon_fault_handler = handle_fault;
	for (i = 0; cmdtable[i].cmd != 0; i++) {
		if (!strcmp(av[1], cmdtable[i].cmd)) {
			int rval = (*cmdtable[i].proc)(ac-1, &av[1], regs);
			xmon_fault_handler = 0;
			return rval;
		}
	}
	printf("unknown \"os\" command\n");
	return 0;
}

#if 0
static struct {
    int         doit;
    int         (*proc)(void);
} checktable[] = {
    { 0, _xos_check_files },
    { 0, _xos_check_inodes },
    { 0, 0 }
};
#define CHECK_FILE      0
#define CHECK_INODE     1

int
xos_check(struct pt_regs *regs) {
    int i;
    int e = 0;
    for (i = 0; checktable[i].proc != 0; i++) {
        if (checktable[i].doit)
            e += (*checktable[i].proc)();
    }
    return e;
}

int
xos_check_opt(int ac, char **av, struct pt_regs *regs) {
    int doit;
    int i;
    if (ac == 1) {
        goto EXIT;
    }
    if (strcmp(av[1], "on") == 0)
        doit = 1;
    else if (strcmp(av[1], "off") == 0)
        doit = 0;
    else {
        goto EXIT;
    }
    if (ac == 2) {
        int i;
        for (i = 0; checktable[i].proc != 0; i++)
            checktable[i].doit = doit;
        return 0;
    }
    i = 2;
    for (; i < ac; i++) {
        if (!strcmp(av[i], "file"))
            checktable[CHECK_FILE].doit = doit;
        else if (!strcmp(av[i], "inode"))
            checktable[CHECK_INODE].doit = doit;
        else
            goto EXIT;
    }
    return 0;
EXIT:
    printf("usage: os check {on|off} [file|inode] \n");
    return 0;
}
#endif

/*
 * For callout debugging
 */
void
_xdebug_illegal(struct pt_regs *regs, unsigned int instr) {
	extern int print_insn_big_powerpc(FILE *, unsigned long, unsigned);

	printk("Illegal instruction: pid %d, \"%s\"\n",
		current->pid, current->comm);
	printk("\t%08lx %08x\t", regs->nip, instr);
	print_insn_big_powerpc(stdout, instr, regs->nip);
	printk("\n");
}
#endif CONFIG_XMON
