/*
 *  linux/drivers/block/ide-tivodma.c
 *
 *  Copyright (c) 1998  Teleworld
 */

#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/timer.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/blkdev.h>
#include <linux/hdreg.h>
#include <linux/pci.h>
#include <linux/bios32.h>

#include <asm/io.h>
#include <asm/dma.h>
#include <asm/ptrace.h>
#include <asm/qdma.h>
#include <asm/px_hw.h>

#include "ide.h"

#define IDESG
#define ULTRA_DMA

#undef DEBUG_DMA
#undef DEBUG_TIVO_CALL
#undef DEBUG
#ifdef IDESG_D
#define DEBUG
#endif

#include <linux/tivo.h>

#include <linux/ide-tivo.h>
#include <linux/tivo-private.h>
#include <linux/tivo-diskmon.h>

#define MIN(a,b) (((a)<(b))?(a):(b))

/*
 * Our Physical Region Descriptor (PRD) table should be large enough
 * to handle the biggest I/O request we are likely to see.  Since requests
 * can have no more than 256 sectors, and since the typical blocksize is
 * two sectors, we could get by with a limit of 128 entries here for the
 * usual worst case.  Most requests seem to include some contiguous blocks,
 * further reducing the number of table entries required.
 *
 * The driver reverts to PIO mode for individual requests that exceed
 * this limit (possible with 512 byte blocksizes, eg. MSDOS f/s), so handling
 * 100% of all crazy scenarios here is not necessary.
 *
 * As it turns out though, we must allocate a full 4KB page for this,
 * so the two PRD tables (ide0 & ide1) will each get half of that,
 * allowing each to have about 256 entries (8 bytes each) from this.
 */
#define DMA_READ        0x0
#define DMA_WRITE       0x10000000
#define MAX_BUF_SIZE    0x0FFFFFFF

static struct wait_queue *ideDMAwait;
static int          IdeChBusyLock  = 0;
static int          ide_dma_ops;
static int          ide_fpga_dma_complete;
static int          ide_drive_complete = 1;
static int          ide_fpga_dma_request_count = 0;
#if 0
static int           lirq;
static unsigned long *ldev_id;
static unsigned long *lregs;
byte                 lstat;
#endif

#define IDE_MAXFRAGS 512

#if 0
typedef void *(TyDmaChT, DmaChReqT *) idedmafunc_t;

static idedmafunc_t ideDmaChFunc = NULL;
#endif
static int onms4 = 0;

void (*idetivo_cmd_hook)(ide_drive_t *drive, unsigned char *cmdcode);
void idetivo_update_stats (enum diskmon newState);

/* TiVo DMA channel data structures. */
static struct ideDmaCh {
  struct _DmaChReqT DmaChReq;
#pragma align 8
  DmaChFragT fragbufs[IDE_MAXFRAGS];
} hwifIdeDmaCh[MAX_HWIFS];

static void ideChWait(void) {
  int icount;
  while(IdeChBusyLock) {
#ifdef DEBUG
    printk("Must wait for the IDE channel!\n");
#endif
    interruptible_sleep_on(&ideDMAwait);
    for(icount=0; icount < 0x1000; icount++);
  }
  IdeChBusyLock = 1;
  return;
}

static void ideChFree(void) {
  wake_up(&ideDMAwait);
  IdeChBusyLock = 0;
  return;
}

void chkIdeChBusyLock(void) {
  if (!IdeChBusyLock)
    return;
  printk("IDE: ChBusyLock error.  Attempting recovery.\n");
  ideChFree();
}

void
IdeDmaChEnqueue(DmaChReqT *req)
{
#if 0
  DmaChEnqueue(tyDmaChIde,req);
#endif
  DmaListChEnqueue(tyDmaChIdeSG,tyDmaChIde,req);
}


int tivoide_dma_free(void) {
   return (ide_fpga_dma_request_count == 0);
}

/*
 * tivo_dma_intr() is the handler for disk interrupts arriving from
 * the drive, during an operation in which DMA data transfer was taking
 * place.
 */
static void
tivo_dma_intr (ide_drive_t *drive) {
   byte stat;
   int  i;
   struct request *rq = HWGROUP(drive)->rq;
   
#ifdef DEBUG
   printk("IDE: in tivo_dma_intr for drive %s\n", drive->name);
#endif
   
   if (ide_fpga_dma_request_count > 1) {
      printk("IDE: %s-done interrupt from drive,\n"
	     "     fpga ide dma engine still has %d requests pending\n",
	     (ide_dma_ops ? "Write":"Read"), ide_fpga_dma_request_count);
   }

   if (drive->dma_running) {
      printk("tivo_dma_intr() called with DMA still running, abort it!\n");
      DmaChStop(tyDmaChIde, 1); /* abort the DMA and clear its interrupt*/
      ide_fpga_dma_complete = 1;
      ide_fpga_dma_request_count = 0;
      ideChFree();
      drive->dma_running = 0;
   }
   
   ide_drive_complete = 1;

   drive->dma_irq_hold = 0;

   idetivo_update_stats(DISKMON_IDLE);

   stat = GET_STAT();                            /* get drive status */
   if (OK_STAT(stat,DRIVE_READY,drive->bad_wstat|DRQ_STAT)) {
      rq = HWGROUP(drive)->rq;
      for (i = rq->nr_sectors; i > 0;) {
	 i -= rq->current_nr_sectors;
	 ide_end_request(1, HWGROUP(drive));
      }
      return;
   }
   /*
    * If we got here, the drive has reported an error status.
    */
   ide_error(drive, "tivo_dma_intr", stat);
   sti();
}

static void
tivo_dma_intr_direct (ide_drive_t *drive) {
  byte stat;
   struct idetivo_request *req; 
#ifdef IDESG_D
   printk("%s\n", __FUNCTION__);
#endif

   req = IDE_TIVO_INFO->cur;
   if (!req) {
      printk(KERN_WARNING "read_intr: no current request! YIKES!"); 
      return; 
   }
   
   if (ide_fpga_dma_request_count > 1) {
      printk("IDE: %s-done interrupt from drive,\n"
	     "     fpga ide dma engine still has %d requests pending\n",
	     (ide_dma_ops ? "Write":"Read"), ide_fpga_dma_request_count);
   }

   if (drive->dma_running) {
      printk("IDE: %s command for sector %d bailed with DMA running\n", 
	     (req->cmd == READ) ? "read" : "write",
	     req->sector);
      DmaChStop(tyDmaChIde, 1); /* abort the DMA and clear its interrupt*/
      ide_fpga_dma_complete = 1;
#ifdef IDESG_D
printk("%s: **0\n",__FUNCTION__);
#endif
      ide_fpga_dma_request_count = 0;   
      ideChFree();
      drive->dma_running = 0;
   }

   drive->dma_running = 0;
   drive->dma_irq_hold = 0;

   stat = GET_STAT();                            /* get drive status */
   if (OK_STAT(stat,DRIVE_READY,drive->bad_wstat|DRQ_STAT)) {
      req->sector_progress += req->chunk_num_sectors;
      req->chunk_sector_progress += req->chunk_num_sectors;
      req->progress += req->chunk_num_sectors * 512; 
      idetivo_release_request(drive, req);
      return;
   }
   /*
    * If we got here, the drive has reported an error status.
    */
   ide_error(drive, "tivo_dma_intr_direct", stat);
   sti();
}

/*
 * FPGA call back functions.
 */
static void
fpga_call_back(DmaChReqT *req, void *ptr) {
   int   dma_stat;
   ide_drive_t *drive = *(ide_drive_t **) ptr;
#ifdef  DEBUG
  printk("IDE: fpga DMA call back, drive %s, ide_fpga_dma_request_count = 0x%x, req=0x%x, ptr=0x%x\n",
         drive->name, ide_fpga_dma_request_count, req, ptr);
#endif

  if(ide_fpga_dma_request_count == 1) {
    /* all DMA request are finished, get DMA status */
#ifdef IDESG_D
printk("%s: **0\n",__FUNCTION__);
#endif
    ide_fpga_dma_request_count = 0;
    dma_stat = DmaChReqIsFinished(req);
  } else {
    /* Decrement the request count and return. */
#ifdef IDESG_D
printk("%s: **decr: %d\n",__FUNCTION__, ide_fpga_dma_request_count);
#endif
    ide_fpga_dma_request_count -= 1;
    if(ide_fpga_dma_request_count <= 0) {
      printk("IDE: DMA in a strange state\n");
#ifdef IDESG_D
      printk("fpga_call_back, count=%d\n", ide_fpga_dma_request_count);
#endif
      ide_fpga_dma_request_count = 0;
      if (!drive->dma_running) {
	printk("FPGA callback with DMA not running - delayed interrupt?\n");
      }
    }
    return;
  }

  /*
   * fpga ide dma engine is in good state, see if we are idle.
   * If the drive had already interrupted to signal completion of
   * the command, free the channel and complete the request.
   */
  if(dma_stat == TRUE) {
#ifdef DEBUG
     printk("End of DMA on drive %s\n", drive->name);
#endif
     ideChFree();
     ide_fpga_dma_complete = 1;
     drive->dma_running = 0;
     if (drive->dma_irq_hold) {
#ifdef DEBUG
	printk("Re-enable masked interrupt\n");
#endif
	enable_irq(HWIF(drive)->irq);
     }
  } else {
      panic("IDE: fpga dma call back, dma engine does not finish the job\n");
     /* XXX TBD - DMA problems.  If there's a command in progress,
      * abort it, reset the drive, and try again.  dcp
      */
  }
  return;
}

static ide_drive_t *fpga_current_drive = NULL;

static DmaChParmsT fpga_call_back_parms = { fpga_call_back,
                                            (void*) &fpga_current_drive };

/*
 * build_dmatable() prepares a dma request.
 * Returns 0 if all went okay, returns 1 otherwise.
 */
static int
build_dmatable (ide_drive_t *drive, unsigned long ops) {
  struct request *rq = HWGROUP(drive)->rq;
  struct buffer_head *bh = rq->bh;
  unsigned long size, addr;
  unsigned int count = 0;
  DmaChReqT *table = (DmaChReqT *)HWIF(drive)->dmatable;

  /* initialize the dma request count to zero.  */
#ifdef IDESG_D
printk("%s: **0\n",__FUNCTION__);
#endif
  ide_fpga_dma_request_count = 0;

  do {
    /*
     * Determine addr and size of next buffer area.  We assume that
     * individual virtual buffers are always composed linearly in
     * physical memory.  For example, we assume that any 8kB buffer
     * is always composed of two adjacent physical 4kB pages rather
     * than two possibly non-adjacent physical 4kB pages.
     */
    if (bh == NULL) {
      /* paging requests have (rq->bh == NULL) */
      addr = virt_to_bus (rq->buffer);
      size = rq->nr_sectors << 9;
    } else {
      /* group sequential buffers into one large buffer */
      addr = virt_to_bus (bh->b_data);
      size = bh->b_size;
      while ((bh = bh->b_reqnext) != NULL) {
        if ((addr + size) != virt_to_bus (bh->b_data)) break;
        size += bh->b_size;
      }
    }

    /*
     * Fill in the dma table, and the buffer
     * size can not exceed the MAX_BUF_SIZE.
     */
    while(size >= (unsigned long)MAX_BUF_SIZE) {
      /* Now we have to chop the buffer into fragments  */
      table->fragp[count].buf = (U32*)addr;
      table->fragp[count].len = (U32)(MAX_BUF_SIZE|ops);
#ifdef  DEBUG
      printk("IDE: frags[%d].buf = 0x%x\n"
             "     frags[%d].len = 0x%x\n", count, addr,
             count, MAX_BUF_SIZE|ops);
#endif
      size -= MAX_BUF_SIZE;
      addr += MAX_BUF_SIZE;
      count += 1;
      if (count >= table->maxFrags) panic("bldfragovflw");
    }
    /* now take care of the left over   */
    table->fragp[count].buf = (U32*)addr;
    table->fragp[count].len = (U32)(size|ops);
#ifdef DEBUG
    printk("IDE: frags[%d].buf = 0x%x\n"
           "     frags[%d].len = 0x%x\n", count, addr,
           count, size|ops);
#endif
    count += 1;
    if (count >= table->maxFrags) panic("bldfragovflw1");
  } while (bh != NULL);

  if (count <= 0) panic("bldfragnocount");
  table->numFrags = count;
  table->next = NULL;
  table->finished = 0;
#if 0
  IdeDmaChFunc(tyDmaChIde,table);
#endif
  if (onms4)
    IdeDmaChEnqueue(table);
  else
    DmaChEnqueue(tyDmaChIde,table);
#ifdef IDESG_D
printk("%s: **incr %d\n", __FUNCTION__,ide_fpga_dma_request_count);
#endif
  ide_fpga_dma_request_count += 1;
  return 0;
}

/*
 * build_dmatable_tivo() prepares a dma request.
 * Returns 0 if all went okay, returns 1 otherwise.
 */
static int
build_dmatable_tivo (ide_drive_t *drive, unsigned long ops) {
   struct idetivo_request *req = IDE_TIVO_INFO->cur;
   struct FsIovec *vec = req->state.top;
   unsigned long bytes = req->chunk_num_sectors * SECTOR_BYTES;
   unsigned long sum_bytes;
   unsigned long size, addr;
   int count = -1;
   DmaChReqT *table = (DmaChReqT *)HWIF(drive)->dmatable;
   unsigned long bcount, nextAddress = 0, hasRoom = 0;
   long int totBytes = 0;

  /* initialize the dma request count to zero.  */
#ifdef IDESG_D
printk("%s: **0\n",__FUNCTION__);
#endif
  ide_fpga_dma_request_count = 0;
#ifdef DEBUG_DMA
   printk("build_dmatable_tivo()\n");
#endif
   
   if (bytes % 512 != 0) {
      printk("AAGH!  build_dma_tivo byte count %ld is not 0 mod 512!\n", bytes);
   }

  sum_bytes = 0; 
   while (sum_bytes < bytes) {
      /*
       * Fill in the dma table, without exceeding the limits.
       */
      addr = virt_to_bus(vec->pb);
      size = MIN((unsigned long) vec->cb, bytes - sum_bytes);
      vec++;
      sum_bytes += size; 
      while (size) {
	 if (!hasRoom || addr != nextAddress) {
	    if (count >= 0) {
	       table->fragp[count].len |= ops;
	    }
	    ++count;
	    if (count >= table->maxFrags) panic("tbldfragovflw");

	    table->fragp[count].buf = (U32*) nextAddress = addr;
	    table->fragp[count].len = 0;;
	    hasRoom = 1;
	 }
	 bcount = MAX_BUF_SIZE - table->fragp[count].len;
	 if (bcount > size) {
	    bcount = size;
	 } else {
	    hasRoom = 0;
	 }
#ifdef DEBUG_DMA
	 printk("DMA %d bytes at 0x%X\n", bcount, addr);
#endif   
	 table->fragp[count].len += bcount;
	 nextAddress += bcount;
	 addr = nextAddress;
	 size -= bcount;
	 totBytes += bcount;
#ifdef DEBUG_DMA
	 if (!hasRoom) printk("This DMA table entry hit the Energy Barrier\n");
#endif   
      }
   }
   /* now take care of the left over   */
   table->fragp[count].len |= ops;
#ifdef DEBUG_DMA
   printk("IDE: frags[%d].buf = 0x%x\n"
	  "     frags[%d].len = 0x%x\n", count, table->fragp[count].buf,
	  count, table->fragp[count].len);
#endif
   count += 1;
   /*
    * Now enqueue the left over frags.
    */
   table->numFrags = count;
   table->next = NULL;
   table->finished = 0;
   if (onms4)
     IdeDmaChEnqueue(table);
   else
     DmaChEnqueue(tyDmaChIde,table);
#ifdef IDESG_D
printk("%s: **incr %d\n",__FUNCTION__,ide_fpga_dma_request_count);
#endif
   ide_fpga_dma_request_count += 1;
   return 0;
}

static int
config_drive_for_dma (ide_drive_t *drive) {
  struct hd_driveid *id = drive->id;

  if ((id->field_valid & 2) && (id->dma_mword & 0x404) == 0x404) {

      drive->using_dma = 1;

#ifdef ULTRA_DMA
      /* If we don't have MS4 or greater there isn't any ultra DMA support */
      if (!ms4probe())
	return 0;

      /* Only support drives with ultra DMA supprot */
      if ((id->dma_ultra & 0x7) != 0x7)
	return 0;;

      /* Set up MS timing registers */
      if (drive->name[2] == 'a') {
	/* Assume we are dealing with the master (unit 0) */
	MsWrite(MS_IDE0_ULTRA_DMA_TIMING_REG, 0x62);
      } else {
	/* Slave drive (unit 1). */
	MsWrite(MS_IDE1_ULTRA_DMA_TIMING_REG, 0x62);
      }
      MsWrite(MS_IDE_ULTRA_DMA_CONTROL_REG, 3);

      /* Enable Ultra DMA on drive */
      OUT_BYTE (drive->select.all, IDE_SELECT_REG);
      OUT_BYTE(drive->ctl,IDE_CONTROL_REG);
      OUT_BYTE(0x42,IDE_NSECTOR_REG);
      OUT_BYTE(0x3,IDE_FEATURE_REG);
      OUT_BYTE(WIN_SETFEATURES, IDE_COMMAND_REG);

      printk("%4s: Ultra DMA enabled\n", drive->name);
#endif
      return 0;
  } else {
    printk("IDE: Can not use DMA for drive %s\n", drive->name);
    return 1;	/* DMA not enabled */
  }
}

/*
 * tivo_ide_dmaproc() initiates/aborts DMA read/write operations on a drive.
 *
 * The caller is assumed to have selected the drive and programmed the drive's
 * sector address using CHS or LBA.  All that remains is to prepare for DMA
 * and then issue the actual read/write DMA/PIO command to the drive.
 *
 * For ATAPI devices, we just prepare for DMA and return. The caller should
 * then issue the packet command to the drive and call us again with
 * ide_dma_begin afterwards.
 *
 * Returns 0 if all went well.
 * Returns 1 if DMA read/write could not be started, in which case
 * the caller should revert to PIO for the current request.
 */
static int
tivo_ide_dmaproc (ide_dma_action_t func, ide_drive_t *drive) {
  unsigned long frdwr = DMA_READ;
  unsigned long reading = 1;
  int           wait_count;
  long int      flags;
  unsigned char cmdcode;

#ifdef DEBUG
  printk("tivo_ide_dmaproc: func %d\n", func);
#endif

   if (func == ide_dma_abort) {
      long int flags;
      save_flags(flags);
      cli();
      (void)DmaChStop(tyDmaChIde, 1);
      ide_fpga_dma_complete = 1;
#ifdef IDESG_D
printk("%s: **0\n", __FUNCTION__);
#endif
      ide_fpga_dma_request_count = 0;   
      ideChFree();
      drive->dma_running = 0;
      if (drive->dma_irq_hold) {
#ifdef DEBUG
	printk("Re-enable masked interrupt\n");
#endif
	 enable_irq(HWIF(drive)->irq);
      }
      restore_flags(flags);
      printk("IDE: tivo DMA engine aborted\n");
      return 0;
   }
   
  /*
   * make sure there is not pending request for current drive
   * so we wont overwritten the dmatable.
   * Do we need this ????  triton chip did not check this at all.
   */
   
  ideChWait();

  /*
   * Now that we have the channel lock its safe to do our things.
   */
  switch (func) {
  case ide_dma_check:
    wait_count = config_drive_for_dma(drive);
    ideChFree();
    return wait_count;
  case ide_dma_write:
    frdwr = DMA_WRITE;
    reading = 0;
    break;
  case ide_dma_read:
    break;
  case ide_dma_status_bad:
    wait_count = DmaChIdle(tyDmaChIde);	/* verify good DMA status */
    ideChFree();
    return wait_count;
  case ide_dma_transferred:
    ideChFree();
    return (0);
  case ide_dma_begin:
    printk("IDE: tivo DMA don't like ide_dma_begin func\n");
    ideChFree();
    return 1;
#if 0
    DmaChStart(tyDmaChIde);             /* begin DMA */
    return 0;
#endif
  default:
    printk("tivo_dmaproc: unsupported func: %d\n", func);
    ideChFree();
    return 1;
  }

   if (ms2probe()) {
      MsWrite(MS2_CONTROL_REGISTER,MsRead(MS2_CONTROL_REGISTER) & ~0x00010000);
      MsWrite(MS2_SYSTEM_REGISTER,MsRead(MS2_SYSTEM_REGISTER) | 0x00000100);
      MsWrite(MS2_IDE_SCRAMBLER_0,0);
      MsWrite(MS2_IDE_SCRAMBLER_1,0);
      MsWrite(MS2_IDE_SCRAMBLER_2,0);
      MsWrite(MS2_IDE_SCRAMBLER_3,0);
   }

   if (IDE_TIVO_INFO && IDE_TIVO_INFO->cur) {
      struct io_scramble scramble;
      if (build_dmatable_tivo (drive, frdwr)) {
	ideChFree();
	return 1;
      }
      scramble = IDE_TIVO_INFO->cur->scramble;
      if (scramble.magic == IO_SCRAMBLE_MAGIC &&
	  scramble.version == IO_SCRAMBLE_VERSION_TIVO1 &&
	  ms2probe()) {
	 scramble.config.wyrd[0] += scramble.recordOffset * 73221;
	 scramble.config.wyrd[0] |= SCRAMBLER_MODE_BLUM_SHUB;
	 scramble.config.wyrd[1] -= scramble.recordOffset;
	 scramble.config.wyrd[1] |= SCRAMBLER_MODE_BLOWFISH;
	 scramble.config.wyrd[2] += scramble.recordOffset * 85565;
	 if (!(scramble.config.wyrd[2] & SCRAMBLER_EXPORT_MASK)) {
	    scramble.config.wyrd[2] |= SCRAMBLER_MODE_AES;
	 }
	 scramble.config.wyrd[3] -= scramble.recordOffset * 27;
	 scramble.config.wyrd[3] |= ~SCRAMBLER_EXPORT_MASK;
	 if (!(scramble.config.wyrd[3] & SCRAMBLER_GOST_MASK)) {
	    scramble.config.wyrd[3] |= SCRAMBLER_MODE_GOST;
	 }
	 MsWrite(MS2_SYSTEM_REGISTER,MsRead(MS2_SYSTEM_REGISTER) & ~0x00000100);
	 MsWrite(MS2_IDE_SCRAMBLER_0,scramble.config.wyrd[0]);
	 MsWrite(MS2_IDE_SCRAMBLER_1,scramble.config.wyrd[1]);
	 MsWrite(MS2_IDE_SCRAMBLER_2,scramble.config.wyrd[2]);
	 MsWrite(MS2_IDE_SCRAMBLER_3,scramble.config.wyrd[3]);
	 MsWrite(MS2_CONTROL_REGISTER,MsRead(MS2_CONTROL_REGISTER) | 0x00010000);
      }
   } else {
      if (build_dmatable (drive, frdwr)) {
	ideChFree();
	return 1;
      }
   }

  /*
   * Both the drive and the DMA engine is ready when we get here.
   * ide Channel will remained locked until the ide dma intr routine
   * get called.
   */
   if (IDE_TIVO_INFO && IDE_TIVO_INFO->cur) {
      ide_set_handler(drive, &tivo_dma_intr_direct, WAIT_CMD);	/* issue cmd to drive */
   } else {
      ide_set_handler(drive, &tivo_dma_intr, WAIT_CMD);	/* issue cmd to drive */
   }
  /* start your engine  */
  ide_dma_ops = frdwr;
   /* avoid race conditions during the following, just to be sure... dcp */
   save_flags(flags);
   cli();
   ide_fpga_dma_complete = 0;
   fpga_current_drive = drive;
   ide_drive_complete = 0;
   drive->dma_running = 1;
#ifdef IDESG
   if (ms4probe()) {
      DmaChStart(tyDmaChIdeSG);
   } else {
#endif
      DmaChStart(tyDmaChIde);
#ifdef IDESG
   }
#endif
   /* get the drive dma going also.  */
   cmdcode = reading ? WIN_READDMA : WIN_WRITEDMA;
   if (idetivo_cmd_hook) {
    (*idetivo_cmd_hook)(drive, &cmdcode);
   }
   OUT_BYTE(cmdcode, IDE_COMMAND_REG);
   restore_flags(flags);
  return 0;
}

#if 0
/*
 * Initialize the dma engine.
 */
void
init_tivo_dma (ide_hwif_t *hwif) {
  static unsigned long dmatable = 0;

  printk("IDE: DMA initialization\n");
  hwif->dma_busy = 0;

  /*
   * allocate space for DMA descriptotr table.
   */
  if (!dmatable)
    dmatable = __get_free_page(GFP_KERNEL);
  if (dmatable) {
    hwif->dmatable = (unsigned long *) dmatable;
    memset(hwif->dmatable, 0, sizeof(DmaChReqT));
    hwif->dmaproc  = &tivo_ide_dmaproc;
  } else {
    printk("Can't get DMA table for TiVo interface\n");
  }
  return;
}
#endif

/*
 * ide_init_triton() prepares the IDE driver for DMA operation.
 * This routine is called once, from ide.c during driver initialization,
 * for each triton chipset which is found (unlikely to be more than one).
 */
void
ide_init_tivo(void) {
  int h;

#ifdef  DEBUG
  printk("IDE: Tivo DMA channel related initialization.\n");
#endif

  /*
   * Setup DMA engine call back function and some dma related flags.
   */
  if(DmaChSetParameters(tyDmaChIde, &fpga_call_back_parms) == -1) {
    panic("IDE: DmaChSetParameters failed.\n");
  } else
    printk("IDE: Setup IDE dma engine call back function completed\n");

  /* Configure DMA for scatter/gather list */
  if (ms4probe()) {
#if 0
    ideDmaChFunc = IdeSGDmaChEnqueue;
#endif
#ifdef IDESG
    onms4 = 1;
#else
    onms4 = 0;
#endif
  } else {
#if 0
    ideDmaChFunc = DmaChEnqueue;
#endif
    onms4 = 0;
  }
  /*
   * Save the dma_base port addr for each interface
   */
  for (h = 0; h < MAX_HWIFS; ++h) {
    ide_hwif_t *hwif = &ide_hwifs[h];
    hwif->chipset = ide_tivodma;
    hwif->dma_busy = 0;
    hwif->dmatable = (unsigned long *) &hwifIdeDmaCh[h].DmaChReq;
printk("set maxFrags and initialize fragp for h=%d\n", h);
    hwifIdeDmaCh[h].DmaChReq.maxFrags = IDE_MAXFRAGS;
    hwifIdeDmaCh[h].DmaChReq.fragp = hwifIdeDmaCh[h].fragbufs;
    hwif->dmaproc  = &tivo_ide_dmaproc;
  }
  IdeChBusyLock = 0;
  ide_dma_ops = 0;
  ide_drive_complete = 1;
  ide_fpga_dma_complete = 1;
}
