
#include<linux/errno.h>
#include<linux/sched.h>
#include<linux/init.h>
#include<linux/module.h>
#include<linux/tivoconfig.h>

typedef struct {
	TivoConfigSelector selector;
	TivoConfigValue value;
} TivoConfig;


static TivoConfig gClassicConfigTable[] = { 
    { kTivoConfigBoardID, kTivoConfigBoardIDClassic },	
    { kTivoConfigMemSize, 0x01000000 }, // 16 MB
    { kTivoConfigI2cType, kTivoConfigI2cTypeSingleMSBus },
    { 0,0 }
};
static TivoConfig gComboConfigTable[] = { 
    { kTivoConfigBoardID, kTivoConfigBoardIDCombo },	
    { kTivoConfigMemSize, 0x02000000 }, // 32 MB
    { kTivoConfigI2cType, kTivoConfigI2cTypeSingleMSBus },
    { 0,0 }
};
static TivoConfig gBrcmUmaConfigTable[] = { 
    { kTivoConfigBoardID, kTivoConfigBoardIDBcmUma },	
    { kTivoConfigMemSize, 0x02000000 }, // 32 MB
    { kTivoConfigI2cType, kTivoConfigI2cTypeDualMSBus },
    { 0,0 }
};

static TivoConfig *gGlobalConfigTable[] = { 
    gClassicConfigTable,
    gComboConfigTable,
    gBrcmUmaConfigTable,
    0
};

EXPORT_SYMBOL(InitTivoConfig);
EXPORT_SYMBOL(GetTivoConfig);
EXPORT_SYMBOL(SetTivoConfig);

TivoConfig gTivoConfig[ kTivoConfigMaxNumSelectors ];
int gTivoConfigCount = 0;

TStatus SetTivoConfig( TivoConfigSelector selector, TivoConfigValue value )
{
    TStatus result = -1;

    long flags;    
  
    save_flags( flags ); 
    
    cli();
    {
	int i;
	for( i = 0; i < gTivoConfigCount; ++i )
	{
	    if( gTivoConfig[ i ].selector == selector )
	    {
		gTivoConfig[ i ].value = value;	    
		result = 0;
		break;	    
	    }	
	}
	if( result != 0 )
	{
	    if( gTivoConfigCount < kTivoConfigMaxNumSelectors )
	    {
		TivoConfig *tivoConfig = &gTivoConfig[ gTivoConfigCount++ ];

		tivoConfig->selector = selector;
		tivoConfig->value = value;	
	
		result = 0;	     	
	    }        
	    else
	    {
		result = -ENOMEM;	
	    }	
	}
    }
    restore_flags( flags ); 

    return result;    
}

TStatus GetTivoConfig( TivoConfigSelector selector, TivoConfigValue *value )
{
    TStatus result = -1;

    int i;
    
    for( i = 0; i < gTivoConfigCount; ++i )
    {
	if( gTivoConfig[ i ].selector == selector )
	{
	    *value = gTivoConfig[ i ].value;	    
	    result = 0;
	    break;	    
	}	
    }
    return result;    
}
 
void __init InitTivoConfig( TivoConfigValue boardID )
{

    TivoConfig *table;
    
    int tableIndex = 0;

    while( ( table = gGlobalConfigTable[ tableIndex++ ] ) )
    {
	if( table[0].value == boardID )
	{
	    int selectorIndex = 0;

	    while( table[ selectorIndex ].selector != 0 )
	    {
		SetTivoConfig( table[ selectorIndex ].selector, table[ selectorIndex ].value );

		selectorIndex += 1;
	    }
	    break;
	}
    }
}


