#ifndef __PPC_MMU_CONTEXT_H
#define __PPC_MMU_CONTEXT_H

#include <asm/ppcops.h>

#ifdef CONFIG_TCD
#define NO_CONTEXT	-1
#define LAST_CONTEXT	255
#else
#define NO_CONTEXT	0
#define LAST_CONTEXT	0xfffff
#endif

extern int next_mmu_context;
extern void mmu_context_overflow(void);

#ifdef CONFIG_TCD
#define set_context(c)	mtspr(SPRN_PID, c)
#define get_context()   mfspr(SPRN_PID)
#else
extern void set_context(int context);
extern int get_context(void);
#endif

/*
 * Get a new mmu context for task tsk if necessary.
 */
#define get_mmu_context(tsk)					\
do { 								\
	struct mm_struct *mm = (tsk)->mm;			\
	if (mm->context == NO_CONTEXT) {			\
		if (next_mmu_context == LAST_CONTEXT)		\
			mmu_context_overflow();			\
		mm->context = next_mmu_context++;\
	}							\
} while (0)

/*
 * Set up the context for a new address space.
 */
#define init_new_context(mm)	if (1) { \
		if (next_mmu_context == LAST_CONTEXT)		\
			mmu_context_overflow();			\
		mm->context = next_mmu_context++; } else ;

/*
 * We're finished using the context for an address space.
 */
#define destroy_context(mm)	do { } while (0)

#endif
